from typing import List, Optional, Union, cast, TypedDict

from .. import exceptions
from ..dependent_resource import DependentResource
from ..pine import PineClient
from ..types import AnyObject
from ..types.models import EnvironmentVariableBase, ServiceType
from ..utils import merge
from ..settings import Settings
from .application import Application


class ServiceNaturalKey(TypedDict):
    application: Union[str, int]
    service_name: str


class Service:
    """
    This class implements service model for balena python SDK.

    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__pine = pine
        self.__application = Application(pine, settings, False)
        self.var = ServiceEnvVariable(pine, self, settings)

    def _get(self, id: int, options: AnyObject = {}):
        service = self.__pine.get({"resource": "service", "id": id, "options": options})

        if service is None:
            raise exceptions.ServiceNotFound(id)

        return service

    def get_all_by_application(self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[ServiceType]:
        """
        Get all services from an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[ServiceType]: service info.
        """

        services = self.__application.get(slug_or_uuid_or_id, {"$select": "service", "$expand": {"service": options}})[
            "service"
        ]

        return cast(List[ServiceType], services)


class ServiceEnvVariable(DependentResource[EnvironmentVariableBase]):
    """
    This class implements Service environment variable model for balena python SDK.

    """

    def __init__(self, pine: PineClient, service: Service, settings: Settings):
        self.__service = service
        self.__application = Application(pine, settings, False)
        super(ServiceEnvVariable, self).__init__(
            "service_environment_variable",
            "name",
            "service",
            self.__get_resource_id,
            pine,
        )

    def __get_resource_id(self, resource_id: Union[int, ServiceNaturalKey]):
        if resource_id is not None and isinstance(resource_id, dict):
            keys = resource_id.keys()
            if len(keys) != 2 or "application" not in keys or "service_name" not in keys:
                raise Exception(f"Unexpected type for id provided in service var model get resource id: {resource_id}")

            service = self.__service.get_all_by_application(
                resource_id["application"], {"$select": "id", "$filter": {"service_name": resource_id["service_name"]}}
            )[0]

            if service is None:
                raise exceptions.ServiceNotFound(resource_id["service_name"])

            return service["id"]
        if not isinstance(resource_id, int):
            raise Exception(f"Unexpected type for id provided in service varModel getResourceId: {resource_id}")
        return self.__service._get(resource_id, {"$select": "id"})["id"]

    def get_all_by_service(
        self, service_id_or_natural_key: Union[int, ServiceNaturalKey], options: AnyObject = {}
    ) -> List[EnvironmentVariableBase]:
        """
        Get all variables for a service.

        Args:
            service_id_or_natural_key (Union[int, ServiceNaturalKey]): service id (number) or appliation-service_name
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: service environment variables.

        Examples:
            >>> balena.models.service.var.get_all_by_service(1234)
            >>> balena.models.service.var.get_all_by_service({'application': 'myorg/myapp', 'service_name': 'service'})
        """
        return super(ServiceEnvVariable, self)._get_all_by_parent(service_id_or_natural_key, options)

    def get_all_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[EnvironmentVariableBase]:
        """
        Get all service variables by application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: application environment variables.

        Examples:
            >>> balena.models.service.var.get_all_by_application(9020)
            >>> balena.models.service.var.get_all_by_application("myorg/myslug")
        """
        app_id = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})["id"]

        return super(ServiceEnvVariable, self)._get_all(
            merge(
                {
                    "$filter": {
                        "service": {
                            "$any": {
                                "$alias": "s",
                                "$expr": {"s": {"application": app_id}},
                            }
                        }
                    },
                    "$orderby": "name asc",
                },
                options,
            )
        )

    def get(self, service_id_or_natural_key: Union[int, ServiceNaturalKey], key: str) -> Optional[str]:
        """
        Get the value of a specific service variable

        Args:
            service_id_or_natural_key (Union[int, ServiceNaturalKey]): service id (number) or appliation-service_name
            key (str): variable name

        Examples:
            >>> balena.models.service.var.get(1234,'test_env4')
            >>> balena.models.service.var.get({'application': 'myorg/myapp', 'service_name': 'service'}, 'VAR')
        """
        return super(ServiceEnvVariable, self)._get(service_id_or_natural_key, key)

    def set(self, service_id_or_natural_key: Union[int, ServiceNaturalKey], key: str, value: str) -> None:
        """
        Set the value of a specific application environment variable.

        Args:
            service_id_or_natural_key (Union[int, ServiceNaturalKey]): service id (number) or appliation-service_name
            key (str): variable name
            value (str): environment variable value.

        Examples:
            >>> balena.models.service.var.set({'application': 'myorg/myapp', 'service_name': 'service'}, 'VAR', 'value')
            >>> balena.models.service.var.set(1234,'test_env4', 'value')
        """
        super(ServiceEnvVariable, self)._set(service_id_or_natural_key, key, value)

    def remove(self, service_id_or_natural_key: Union[int, ServiceNaturalKey], key: str) -> None:
        """
        Clear the value of a specific service variable

        Args:
            service_id_or_natural_key (Union[int, ServiceNaturalKey]): service id (number) or appliation-service_name
            key (str): variable name

        Examples:
            >>> balena.models.service.var.remove({'application': 'myorg/myapp', 'service_name': 'service'}, 'VAR')
            >>> balena.models.service.var.remove(1234,'test_env4')
        """
        super(ServiceEnvVariable, self)._remove(service_id_or_natural_key, key)



import re
from collections import defaultdict
from typing import Any, Dict, List, Literal, Optional, TypedDict, Union
from typing_extensions import NotRequired

from semver.version import Version

from .. import exceptions
from ..balena_auth import request
from ..hup import get_hup_action_type
from ..pine import PineClient
from ..types import AnyObject
from ..utils import compare, merge, normalize_balena_semver
from ..settings import Settings
from .application import Application
from .device_type import DeviceType


class DownloadConfig(TypedDict):
    developmentMode: NotRequired[bool]
    appUpdatePollInterval: NotRequired[int]
    network: NotRequired[Literal["ethernet", "wifi"]]
    wifiKey: NotRequired[str]
    wifiSsid: NotRequired[str]
    appId: NotRequired[int]
    fileType: NotRequired[Union[Literal[".img"], Literal[".zip"], Literal[".gz"]]]
    imageType: NotRequired[Union[Literal["raw"], Literal["flasher"]]]


class ImgConfigOptions(TypedDict, total=False):
    network: Optional[Literal["ethernet", "wifi"]]
    appUpdatePollInterval: Optional[int]
    provisioningKeyName: Optional[str]
    provisioningKeyExpiryDate: Optional[str]
    wifiKey: Optional[str]
    wifiSsid: Optional[str]
    ip: Optional[str]
    gateway: Optional[str]
    netmask: Optional[str]
    deviceType: Optional[str]
    version: str
    developmentMode: Optional[bool]


NETWORK_WIFI = "wifi"
NETWORK_ETHERNET = "ethernet"

NETWORK_TYPES = [NETWORK_WIFI, NETWORK_ETHERNET]

ARCH_COMPATIBILITY_MAP = {"aarch64": ["armv7hf", "rpi"], "armv7hf": ["rpi"]}

VERSION_RANGE_CHAR_LIST = ["x", "X", "*"]


def cmp_to_key(mycmp):
    "Convert a cmp= function into a key= function"

    class K:
        def __init__(self, obj, *args):
            self.obj = obj

        def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0

        def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0

        def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0

        def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0

        def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0

        def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0

    return K


def get_rev_version(semver_version):
    if semver_version and semver_version.build:
        rev = 0

        for metadata_part in semver_version.build.split("."):
            match = re.match(r"rev(\d+)", metadata_part)
            if match:
                rev = int(match.groups()[0])
                break

        return rev

    return 0


def is_development_version(semver_version):
    if semver_version and semver_version.build:
        return "dev" in semver_version.build

    return False


def compare_balena_version(version_a, version_b):
    """
    Based on https://github.com/balena-io-modules/balena-semver#compare

    """

    if not version_a:
        return 0 if not version_b else -1

    if not version_b:
        return 1

    normalized_version_a = normalize_balena_semver(version_a)
    normalized_version_b = normalize_balena_semver(version_b)

    is_valid_semver_version_a = Version.is_valid(normalized_version_a)
    is_valid_semver_version_b = Version.is_valid(normalized_version_b)

    if not is_valid_semver_version_a or not is_valid_semver_version_b:
        if is_valid_semver_version_a:
            # verison b not valid semver
            return 1

        if is_valid_semver_version_b:
            # version a not valid semver
            return -1

        return compare(is_valid_semver_version_a, is_valid_semver_version_b)

    version_a_semver_obj = Version.parse(normalized_version_a)
    version_b_semver_obj = Version.parse(normalized_version_b)

    semver_compare_result = version_a_semver_obj.compare(version_b_semver_obj)

    if semver_compare_result != 0:
        return semver_compare_result

    rev_result = compare(
        get_rev_version(version_a_semver_obj),
        get_rev_version(version_b_semver_obj),
    )

    if rev_result != 0:
        return rev_result

    dev_result = compare(
        is_development_version(version_a_semver_obj),
        is_development_version(version_b_semver_obj),
    )

    if dev_result != 0:
        return dev_result

    # We can ignore the localeCompare since this is only for balena OS version.
    return compare(normalized_version_a, normalized_version_b)


def sort_version(x, y):
    """
    Sort returned objects from device_os.get_supported_versions method.
    """

    return compare_balena_version(x["raw_version"], y["raw_version"])


def bsemver_match_range(version, version_range):
    if Version.is_valid(version):
        try:
            parsed_version = Version.parse(version)
            if Version.is_valid(version_range):
                if version_range and parsed_version.match(f">={version_range}"):
                    return True
            else:
                if version_range[-1] in VERSION_RANGE_CHAR_LIST:
                    # version range contains 'x', 'X' or '*'
                    min_ver = f"{version_range[:-1]}0"
                    max_ver = str(Version.parse(min_ver).next_version("minor"))
                    if parsed_version.match(f">={min_ver}") and parsed_version.match(f"<{max_ver}"):
                        return True
        except Exception:
            return False
    return False


class DeviceOs:
    """
    This class implements device os model for balena python SDK.

    """

    OS_UPDATE_ACTION_NAME = "resinhup"
    RELEASE_POLICY_TAG_NAME = "release-policy"
    ESR_NEXT_TAG_NAME = "esr-next"
    ESR_CURRENT_TAG_NAME = "esr-current"
    ESR_SUNSET_TAG_NAME = "esr-sunset"
    VARIANT_TAG_NAME = "variant"
    VERSION_TAG_NAME = "version"
    BASED_ON_VERSION_TAG_NAME = "meta-balena-base"
    OS_TYPES = {"default": "default", "esr": "esr"}
    OS_VARIANTS = {"production": "prod", "development": "dev"}

    def __init__(self, pine: PineClient, settings: Settings):
        self.__pine = pine
        self.__settings = settings
        self.__device_type = DeviceType(pine, settings)
        self.__application = Application(pine, settings, False)

    def get_available_os_versions(self, device_type: Union[str, List[str]]):
        """
        Get the supported OS versions for the provided device type(s)

        Args:
            device_type (Union[str, List[str]]): device type slug.

        Returns:
            list: balenaOS versions.

        """
        single_device_type = isinstance(device_type, str)
        device_types = device_type if isinstance(device_type, list) else [device_type]
        versions_by_dt = self.__get_all_os_versions(device_types, True)

        if single_device_type:
            return versions_by_dt.get(device_type, [])

        return versions_by_dt

    def get_all_os_versions(self, device_type: Union[str, List[str]], options: AnyObject = {}):
        """
        Get all OS versions for the provided device type(s), inlcuding invalidated ones

        Args:
            device_type (Union[str, List[str]]): device type slug.
            options (AnyObject): extra pine options to use

        Returns:
            list: balenaOS versions.

        """
        single_device_type = isinstance(device_type, str)
        device_types = device_type if isinstance(device_type, list) else [device_type]

        if options == {}:
            versions_by_dt = self.__get_all_os_versions(device_types)
        else:
            hostapps = self.__get_os_versions(device_types, options)
            versions_by_dt = self.__transform_host_apps(hostapps)

        if single_device_type:
            return versions_by_dt.get(device_type, [])

        return versions_by_dt

    def get_download_size(self, device_type: str, version: str = "latest") -> float:
        """
        Get OS download size estimate. Currently only the raw (uncompressed) size is reported.

        Args:
            device_type (str): device type slug.
            version (str): semver-compatible version or 'latest', defaults to 'latest'.
            * The version **must** be the exact version number.

        Returns:
            float: OS image download size, in bytes.
        """
        slug = self.__device_type.get(device_type, {"$select": "slug"})["slug"]
        return float(
            request(
                method="GET",
                settings=self.__settings,
                path=f"/device-types/v1/{slug}/images/{version}/download-size",
            )["size"]
        )

    def get_max_satisfying_version(
        self,
        device_type: str,
        version_or_range: str = "latest",
        os_type: Optional[Literal["default", "esr"]] = None,
    ) -> Optional[str]:
        """
        Get OS download size estimate. Currently only the raw (uncompressed) size is reported.

        Args:
            device_type (str): device type slug.
            version_or_range (str): can be one of the exact version number,
            in which case it is returned if the version is supported,
            or `None` is returned otherwise,
            * a [semver](https://www.npmjs.com/package/semver)-compatible
            range specification, in which case the most recent satisfying version is returned
            if it exists, or `None` is returned,
            `'latest'` in which case the most recent version is returned, including pre-releases,
            `'recommended'` in which case the recommended version is returned, i.e. the most
            recent version excluding pre-releases, which can be `None` if only pre-release versions
            are available,
            `'default'` in which case the recommended version is returned if available,
            or `latest` is returned otherwise.
            Defaults to `'latest'`
            os_type (Optional[Literal["default", "esr"]]): The used OS type.

        Returns:
            float: OS image download size, in bytes.
        """

        slug = self.__device_type.get(device_type, {"$select": "slug"})["slug"]
        os_versions: List[Any] = self.get_available_os_versions(slug)  # type: ignore

        if os_type is not None:
            os_versions = [osv for osv in os_versions if osv["os_type"] == os_type]

        if version_or_range == "recommended":
            return next(
                (v.get("raw_version") for v in os_versions if v.get("is_recommended")),
                None,
            )

        if version_or_range == "latest":
            return os_versions[0].get("raw_version") if os_versions else None

        if version_or_range == "default":
            return next(
                (v["raw_version"] for v in os_versions if v.get("is_recommended")),
                os_versions[0]["raw_version"] if os_versions else None,
            )

        versions = [v.get("raw_version") for v in os_versions]

        if version_or_range in versions:
            return version_or_range

        parsed_versions = [Version.parse(v) for v in versions if Version.is_valid(v)]
        satisfying_versions = [v for v in parsed_versions if v.match(version_or_range)]
        if not satisfying_versions:
            return None
        return str(max(satisfying_versions))

    def download(
        self,
        device_type: str,
        version: str = "latest",
        options: DownloadConfig = {},
    ):
        """
        Get OS download size estimate. Currently only the raw (uncompressed) size is reported.

        Args:
            device_type (str): device type slug.
            version (str): semver-compatible version or 'latest', defaults to 'latest'.
            * The version **must** be the exact version number.
            options (DownloadConfig): OS configuration options to use.

        Returns:
            float: OS image download size, in bytes.

        Example:
            >>> with b.models.device_os.download("raspberrypi3") as stream:
            ...    stream.raise_for_status()
            ...    with open("my-image-filename", "wb") as f:
            ...        for chunk in stream.iter_content(chunk_size=8192):
            ...            f.write(chunk)
        """
        slug = self.__device_type.get(device_type, {"$select": "slug"})["slug"]

        if version == "latest":
            versions = [v for v in self.get_available_os_versions(slug) if v["os_type"] in self.OS_TYPES["default"]]
            version = next(
                (v["raw_version"] for v in versions if v.get("is_recommended")),
                versions[0]["raw_version"] if versions else None,  # type: ignore
            )
        else:
            version = normalize_balena_semver(version)

        return request(
            method="GET",
            settings=self.__settings,
            path="/download",
            qs={**options, "deviceType": slug, "version": version},
            return_raw=True,
            stream=True,
        )

    def get_config(self, slug_or_uuid_or_id: Union[str, int], options: ImgConfigOptions):
        """
        Download application config.json.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (ImgConfigOptions): OS configuration dict to use. The available options
            are listed below:
                network (Optional[Literal["ethernet", "wifi"]]): The network type that
                the device will use, one of 'ethernet' or 'wifi'.
                appUpdatePollInterval (Optional[int]): How often the OS checks for updates, in minutes.
                provisioningKeyName (Optional[str]): Name assigned to API key
                provisioningKeyExpiryDate (Optional[str]): Expiry Date assigned to API key
                wifiKey (Optional[str]): The key for the wifi network the device will connect to.
                wifiSsid (Optional[str]): The ssid for the wifi network the device will connect to.
                ip (Optional[str]): static ip address
                gateway (Optional[str]): static ip gateway.
                netmask (Optional[str]): static ip netmask.
                deviceType (Optional[str]): The device type.
                version (str): Required: the OS version of the image.
                developmentMode (Optional[bool]): If the device should be in development mode.

        Returns:
            dict: application config.json content.

        Raises:
            ApplicationNotFound: if application couldn't be found.

        """

        if options.get("version") is None:
            raise Exception("An OS version is required when calling os.getConfig")

        options["network"] = options.get("network", "ethernet")
        app_id = self.__application.get_id(slug_or_uuid_or_id)

        try:
            return request(
                method="POST",
                settings=self.__settings,
                path="/download-config",
                body={**options, "appId": app_id},
            )
        except exceptions.RequestError as e:
            if e.status_code == 404:
                raise exceptions.ApplicationNotFound(slug_or_uuid_or_id)
            raise e

    def is_supported_os_update(self, device_type: str, current_version: str, target_version: str) -> bool:
        """
         Returns the supported OS update targets for the provided device type.

        Args:
            device_type (str): device type slug.
            current_version (str): emver-compatible version for the starting OS version
            target_version (str): semver-compatible version for the target OS version
        """
        try:
            action_type = get_hup_action_type(device_type, current_version, target_version)
            if action_type is not None:
                return True
            return False
        except exceptions.OsUpdateError:
            return False

    def get_supported_os_update_versions(self, device_type: str, current_version: str):
        """
        Get OS supported versions.

        Args:
            device_type (str): device type slug.
            current_version (str): device type slug.
        """
        slug = self.__device_type.get(device_type, {"$select": "slug"})["slug"]
        all_versions = self.get_available_os_versions(slug)
        all_versions = [v.get("raw_version") for v in all_versions if v.get("os_type") == self.OS_TYPES["default"]]

        current = next(
            (v for v in all_versions if str(Version.parse(v)) == str(Version.parse(current_version))),
            None,
        )

        versions = [v for v in all_versions if self.is_supported_os_update(device_type, current_version, v)]
        recommended = next((v for v in versions if not Version.parse(v).prerelease), None)

        return {"versions": versions, "recommended": recommended, "current": current}

    def is_architecture_compatible_with(self, os_architecture: str, application_architecture: str):
        """
        Returns whether the specified OS architecture is compatible with the target architecture.

        Args:
            os_architecture (str): The OS's architecture as specified in its device type.
            application_architecture (str): The application's architecture as specified in its device type.

        Returns:
            bool: Whether the specified OS architecture is capable of
                  running applications build for the target architecture.
        """

        if os_architecture != application_architecture:
            if (
                os_architecture in ARCH_COMPATIBILITY_MAP
                and application_architecture in ARCH_COMPATIBILITY_MAP[os_architecture]
            ):
                return True
            return False

        return True

    def __tags_to_dict(self, tags: List[Any]) -> Dict[str, str]:
        tag_map = {}

        for app_tag in tags:
            tag_map[app_tag["tag_key"]] = app_tag["value"]

        return tag_map

    def __get_os_app_tags(self, app_tags: List[Any]):
        tag_map = self.__tags_to_dict(app_tags)
        return {
            "os_type": tag_map.get(self.RELEASE_POLICY_TAG_NAME, self.OS_TYPES["default"]),
            "next_line_version_range": tag_map.get(self.ESR_NEXT_TAG_NAME, ""),
            "current_line_version_range": tag_map.get(self.ESR_CURRENT_TAG_NAME, ""),
            "sunset_line_version_range": tag_map.get(self.ESR_SUNSET_TAG_NAME, ""),
        }

    def __get_os_versions(self, device_types: List[str], options: AnyObject = {}):
        owns_release = merge(
            {
                "$select": [
                    "id",
                    "known_issue_list",
                    "raw_version",
                    "variant",
                    "phase",
                ],
                "$expand": {"release_tag": {"$select": ["tag_key", "value"]}},
            },
            options,
        )

        return self.__pine.get(
            {
                "resource": "application",
                "options": {
                    "$select": "is_for__device_type",
                    "$expand": {
                        "application_tag": {"$select": ["tag_key", "value"]},
                        "is_for__device_type": {"$select": "slug"},
                        "owns__release": owns_release,
                    },
                    "$filter": {
                        "is_host": True,
                        "is_for__device_type": {
                            "$any": {
                                "$alias": "dt",
                                "$expr": {"dt": {"slug": {"$in": device_types}}},
                            }
                        },
                    },
                },
            }
        )

    def __get_all_os_versions(self, device_types: List[str], listed_by_default: bool = False):
        listed_by_filter = (
            {
                "$filter": {
                    "is_final": True,
                    "is_invalidated": False,
                    "status": "success",
                }
            }
            if listed_by_default
            else {}
        )
        hostapps = self.__get_os_versions(device_types, listed_by_filter)
        return self.__transform_host_apps(hostapps)

    # TODO: Drop this method & just use `release.phase` in the next major
    def __get_os_version_release_line(self, phase: Optional[str], version: str, app_tags: Dict[str, str]):
        if not phase:
            # All patches belong to the same line.
            if bsemver_match_range(version, app_tags["next_line_version_range"]):
                return "next"

            if bsemver_match_range(version, app_tags["current_line_version_range"]):
                return "current"

            if bsemver_match_range(version, app_tags["sunset_line_version_range"]):
                return "sunset"

            if app_tags["os_type"].lower() == self.OS_TYPES["esr"]:
                return "outdated"

        if phase == "end-of-life":
            return "outdated"

        return phase

    def __get_os_versions_from_releases(self, releases: List[Any], app_tags: Any):
        os_variant_names = self.OS_VARIANTS.keys()
        releases_with_os_versions = []

        for release in releases:
            tag_map = self.__tags_to_dict(release.get("release_tag", []))

            release_semver_obj = (
                Version.parse(release["raw_version"]) if not release["raw_version"].startswith("0.0.0") else None
            )

            variant = release.get("variant")
            if variant == "":
                variant = None

            stripped_version = None
            if release_semver_obj is None:
                full_variant_name = tag_map.get(self.VARIANT_TAG_NAME)
                if isinstance(full_variant_name, str):
                    if full_variant_name in os_variant_names:
                        variant = self.OS_VARIANTS[full_variant_name]
                    else:
                        variant = full_variant_name
                else:
                    variant = None
                stripped_version = tag_map.get(self.VERSION_TAG_NAME, "")

                release["raw_version"] = ".".join([x for x in [stripped_version, variant] if x])
            else:
                version = str(release_semver_obj.finalize_version())
                builds = (release_semver_obj.build or "").split(".")
                non_variant_build_parts = ".".join([build for build in builds if build != release["variant"]])

                stripped_version = "+".join([x for x in [version, non_variant_build_parts] if x])

            # TODO: Drop this call & just use `release.phase` in the next major
            line = self.__get_os_version_release_line(release["phase"], stripped_version, app_tags)

            release.update(
                {
                    "variant": variant,
                    "os_type": app_tags.get("os_type"),
                    "line": line,
                    "raw_version": release["raw_version"],
                    "stripped_version": stripped_version,
                    "based_on_version": tag_map.get(self.BASED_ON_VERSION_TAG_NAME, stripped_version),
                }
            )
            releases_with_os_versions.append(release)

        return releases_with_os_versions

    def __transform_host_apps(self, host_apps):
        os_versions_by_device_type = defaultdict(list)

        for host_app in host_apps:
            host_app_device_type = host_app.get("is_for__device_type", [{}])[0].get("slug")
            if host_app_device_type is None:
                continue

            app_tags = self.__get_os_app_tags(host_app.get("application_tag", []))

            os_versions_by_device_type[host_app_device_type] += self.__get_os_versions_from_releases(
                host_app.get("owns__release", []), app_tags
            )

        for device_type in os_versions_by_device_type:
            os_versions_by_device_type[device_type].sort(reverse=True, key=cmp_to_key(sort_version))
            recommended_per_os_type: Dict[str, bool] = {}

            for version in os_versions_by_device_type[device_type]:
                if version["os_type"] not in recommended_per_os_type:
                    if (
                        version["variant"] != "dev"
                        and not version["known_issue_list"]
                        and not Version.parse(version["raw_version"]).prerelease
                    ):
                        additional_format = (
                            f" ({version['line']}, recommended)" if version.get("line") else " (recommended)"
                        )
                        version["is_recommended"] = True
                        version["formatted_version"] = f"v{version['stripped_version']}{additional_format}"
                        recommended_per_os_type[version["os_type"]] = True

        return os_versions_by_device_type



from .balena_auth import request
from .exceptions import BuilderRequestError
from .settings import Settings


def build_from_url(owner: str, app_name: str, url: str, flatten_tarball: bool, settings: Settings) -> int:
    res = request(
        method="post",
        path=f"/v3/buildFromUrl?headless=true&owner={owner}&app={app_name}",
        settings=settings,
        endpoint=settings.get("builder_url"),
        body={"url": url, "shouldFlatten": flatten_tarball},
    )
    if not res.get("started"):
        raise BuilderRequestError(res.get("message"))

    return res["releaseId"]



from typing import Union, TypedDict, Literal, Optional, List
from .organization import Organization
from ..pine import PineClient
from ..settings import Settings
from ..balena_auth import request


class BillingAccountAddressInfo(TypedDict):
    address1: str
    address2: str
    city: str
    state: str
    zip: str
    country: str
    phone: str


class BillingAccountInfo(TypedDict):
    account_state: str
    first_name: str
    last_name: str
    company_name: str
    email: str
    cc_emails: str
    vat_number: str
    address: BillingAccountAddressInfo


BillingInfoType = Literal["bank_account", "credit_card", "paypal"]


class BillingInfo(TypedDict):
    full_name: str
    first_name: str
    last_name: str
    company: str
    vat_number: str
    address1: str
    address2: str
    city: str
    state: str
    zip: str
    country: str
    phone: str

    type: Optional[BillingInfoType]


class CardBillingInfo(BillingInfo):
    card_type: str
    year: str
    month: str
    first_one: str
    last_four: str


class BankAccountBillingInfo(BillingInfo):
    account_type: str
    last_four: str
    name_on_account: str
    routing_number: str


TokenBillingSubmitInfo = TypedDict("TokenBillingSubmitInfo", {"token_id": str, "g-recaptcha-response": Optional[str]})


class Charge(TypedDict):
    itemType: str
    name: str
    code: str
    unitCostCents: str
    quantity: str
    isQuantifiable: Optional[bool]


class BillingPlanBillingInfo(TypedDict):
    currency: str
    totalCostCents: str
    charges: List[Charge]


class Addon(TypedDict):
    code: str
    unitCostCents: Optional[str]
    quantity: Optional[str]


class BillingAddonPlanInfo(TypedDict):
    code: str
    currentPeriodEndDate: Optional[str]
    billing: BillingPlanBillingInfo
    addOns: List[Addon]


SupportInfo = TypedDict("Support", {"name": str, "title": str})


class BillingPlanInfo(TypedDict):
    name: str
    title: str
    code: str
    tier: str
    currentPeriodEndDate: Optional[str]
    intervalUnit: Optional[str]
    intervalLength: Optional[str]
    addonPlan: Optional[BillingAddonPlanInfo]
    billing: BillingPlanBillingInfo
    support: SupportInfo


class InvoiceInfo(TypedDict):
    closed_at: str
    created_at: str
    due_on: str
    currency: str
    invoice_number: str
    subtotal_in_cents: str
    total_in_cents: str
    uuid: str
    state: Literal["pending", "paid", "failed", "past_due"]


class PlanChangeOptions(TypedDict):
    tier: str
    cycle: Literal["monthly", "annual"]
    planChangeReason: Optional[str]


class UpdateAccountBody(TypedDict, total=False):
    email: str
    cc_emails: str


class Billing:
    """
    This class implements billing model for balena python SDK.
    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__settings = settings
        self.__organization = Organization(pine, settings)

    def __get_org_id(self, organization: Union[str, int]) -> int:
        return self.__organization.get(organization, {"$select": "id"})["id"]

    def get_account(self, organization: Union[str, int]) -> BillingAccountInfo:
        """
        Get the user's billing account

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.

        Returns:
            BillingAccountInfo: billing account.

        Examples:
            >>> balena.models.billing.get_account('myorghandle')
        """
        org_id = self.__get_org_id(organization)
        return request(method="GET", settings=self.__settings, path=f"/billing/v1/account/{org_id}")

    def get_plan(self, organization: Union[str, int]) -> BillingPlanInfo:
        """
        Get the current billing plan

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.

        Returns:
            BillingPlanInfo: billing account.

        Examples:
            >>> balena.models.billing.get_plan('myorghandle')
        """
        org_id = self.__get_org_id(organization)
        return request(method="GET", settings=self.__settings, path=f"/billing/v1/account/{org_id}/plan")

    def get_billing_info(self, organization: Union[str, int]) -> BillingInfo:
        """
        Get the current billing information

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.

        Returns:
            BillingInfo: billing information.

        Examples:
            >>> balena.models.billing.get_billing_info('myorghandle')
        """
        org_id = self.__get_org_id(organization)
        return request(method="GET", settings=self.__settings, path=f"/billing/v1/account/{org_id}/info")

    def update_billing_info(self, organization: Union[str, int], billing_info: TokenBillingSubmitInfo) -> BillingInfo:
        """
        Updates the current billing information

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.
            billing_info (TokenBillingSubmitInfo): a dict containing a billing info token_id

        Returns:
            BillingInfo: billing information.

        Examples:
            >>> balena.models.billing.update_billing_info('myorghandle', { token_id: 'xxxxxxx' })
        """
        org_id = self.__get_org_id(organization)
        return request(
            method="PATCH", settings=self.__settings, path=f"/billing/v1/account/{org_id}/info", body=billing_info
        )

    def update_account_info(self, organization: Union[str, int], account_info: UpdateAccountBody) -> None:
        """
        Update the current billing account information

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.
            account_info (UpdateAccountBody): a dict containing billing account info

        Examples:
            >>> balena.models.billing.update_account_info('myorghandle', { email: 'hello@balena.io' })
        """
        org_id = self.__get_org_id(organization)
        return request(
            method="PATCH", settings=self.__settings, path=f"/billing/v1/account/{org_id}", body=account_info
        )

    def change_plan(self, organization: Union[str, int], plan_change_options: PlanChangeOptions) -> None:
        """
        Change the current billing plan

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.
            plan_change_options (PlanChangeOptions): a dict containing billing account info

        Examples:
            >>> balena.models.billing.change_plan('myorghandle', { billingCode: 'prototype-v2', cycle: 'annual' })
        """
        org_id = self.__get_org_id(organization)
        body = {**plan_change_options, "annual": plan_change_options["cycle"] == "annual"}
        return request(method="PATCH", settings=self.__settings, path=f"/billing/v1/account/{org_id}", body=body)

    def get_invoices(self, organization: Union[str, int]) -> List[InvoiceInfo]:
        """
        Get the available invoices

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.

        Examples:
            >>> balena.models.billing.get_invoices('myorghandle')
        """
        org_id = self.__get_org_id(organization)
        return request(method="GET", settings=self.__settings, path=f"/billing/v1/account/{org_id}/invoices")

    def download_invoice(self, organization: Union[str, int], invoice_number: Union[str, int]):
        """
        Download a specific invoice

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.
            invoice_number (Union[str, int]): an invoice number (or the number inside a string)

        Examples:
            >>> with b.models.billing.download_invoice("myorg", "0000") as stream:
            ...    stream.raise_for_status()
            ...    with open("myinvoice.pdf", "wb") as f:
            ...        for chunk in stream.iter_content(chunk_size=8192):
            ...            f.write(chunk)
        """
        org_id = self.__get_org_id(organization)
        return request(
            method="GET",
            path=f"/billing/v1/account/{org_id}/invoices/{invoice_number}/download",
            settings=self.__settings,
            stream=True,
            return_raw=True,
        )



from typing import List, Optional, Union, Dict
import io
from .. import exceptions
from ..balena_auth import request
from ..dependent_resource import DependentResource
from ..pine import PineClient
from ..types import AnyObject, ResourceKey
from ..types.models import (
    OrganizationInviteType,
    OrganizationMembershipTagType,
    OrganizationMembershipType,
    OrganizationType,
)
from ..utils import is_id, merge
from ..settings import Settings


class Organization:
    """
    This class implements organization model for balena python SDK.

    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__pine = pine
        self.invite = OrganizationInvite(pine, self, settings)
        self.membership = OrganizationMembership(pine, self)

    def create(
        self, name: str, handle: Optional[str] = None, logo_image: Optional[io.BufferedReader] = None
    ) -> OrganizationType:
        """
        Creates a new organization.

        Args:
            name (str): the name of the organization that will be created.
            handle (Optional[str]): The handle of the organization that will be created.
            logo_image (Optional[io.BufferedReader]): The organization logo to be used.

        Returns:
            dict: organization info.

        Examples:
            >>> balena.models.organization.create('My Org', 'test_org')
            >>> with open('mypath/myfile.png', 'rb') as f:
            >>>     org = sdk.models.organization.create("my-name", None, f)
        """

        data: Dict[str, Union[str, io.BufferedReader]] = {"name": name}
        if handle is not None:
            data["handle"] = handle

        if logo_image is not None:
            data["logo_image"] = logo_image

        return self.__pine.post({"resource": "organization", "body": data})

    def get_all(self, options: AnyObject = {}) -> List[OrganizationType]:
        """
        Get all organizations.

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[OrganizationType]: list contains information of organizations.

        Examples:
            >>> balena.models.organization.get_all()
        """

        return self.__pine.get(
            {
                "resource": "organization",
                "options": merge({"$orderby": "name asc"}, options),
            }
        )

    def get(self, handle_or_id: Union[str, int], options: AnyObject = {}) -> OrganizationType:
        """
        Get a single organization.

        Args:
            handle_or_id (str): organization handle (string) or id (number).
            options (AnyObject): extra pine options to use

        Returns:
            dict: organization info.

        Raises:
            OrganizationNotFound: if organization couldn't be found.

        Examples:
            >>> balena.models.organization.get(26474)
            >>> balena.models.organization.get('myorg')
        """

        if handle_or_id is None:
            raise exceptions.InvalidParameter("handle_or_id", handle_or_id)

        org = self.__pine.get(
            {
                "resource": "organization",
                "id": handle_or_id if is_id(handle_or_id) else {"handle": handle_or_id},
                "options": options,
            }
        )

        if org is None:
            raise exceptions.OrganizationNotFound(handle_or_id)

        return org

    def remove(self, handle_or_id: Union[str, int]) -> None:
        """
        Remove an organization.

        Args:
            handle_or_id (str): organization handle (string) or id (number).

        Examples:
            >>> balena.models.organization.remove(148003)
        """
        org_id = self.get(handle_or_id, {"$select": "id"})["id"]
        self.__pine.delete({"resource": "organization", "id": org_id})


class OrganizationInvite:
    """
    This class implements organization invite model for balena python SDK.

    """

    def __init__(self, pine: PineClient, organization: Organization, settings: Settings):
        self.__pine = pine
        self.__organization = organization
        self.__settings = settings
        self.RESOURCE = "invitee__is_invited_to__organization"

    def get_all(self, options: AnyObject = {}) -> List[OrganizationInviteType]:
        """
        Get all invites.

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[OrganizationInviteType]: list contains info of invites.

        Examples:
            >>> balena.models.organization.invite.get_all()
        """
        return self.__pine.get({"resource": self.RESOURCE, "options": options})

    def get_all_by_organization(
        self, handle_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[OrganizationInviteType]:
        """
        Get all invites by organization.

        Args:
            handle_or_id (Union[str, int]): organization handle (string), or id (number).
            options (AnyObject): extra pine options to use.

        Returns:
            List[OrganizationInviteType]: list contains info of invites.

        Examples:
            >>> balena.models.organization.invite.get_all_by_organization(26474)
        """
        org_id = self.__organization.get(handle_or_id, {"$select": "id"})["id"]
        return self.get_all(merge({"$filter": {"is_invited_to__organization": org_id}}, options))

    def create(
        self,
        handle_or_id: Union[str, int],
        invitee: str,
        role_name: Optional[str] = None,
        message: Optional[str] = None,
    ) -> OrganizationInviteType:
        """
        Creates a new invite for an organization.

        Args:
            handle_or_id (Union[str, int]): organization handle (string), or id (number).
            invitee (str): the email/balena_username of the invitee.
            role_name (Optional[str]): the role name to be granted to the invitee.
            message (Optional[str]): the message to send along with the invite.

        Returns:
            OrganizationInviteType: organization invite.

        Examples:
            >>> balena.models.organization.invite.create(26474, 'invitee@example.org', 'member', 'Test invite')
        """

        org_id = self.__organization.get(handle_or_id, {"$select": "id"})["id"]
        roles = None
        if role_name is not None:
            roles = self.__pine.get(
                {
                    "resource": "organization_membership_role",
                    "options": {
                        "$top": 1,
                        "$select": ["id"],
                        "$filter": {"name": role_name},
                    },
                }
            )
        body = {
            "invitee": invitee,
            "is_invited_to__organization": org_id,
            "message": message,
        }
        if roles is not None:
            if len(roles) == 0 and role_name is not None:
                raise exceptions.BalenaOrganizationMembershipRoleNotFound(role_name)
            body["organization_membership_role"] = roles[0].get("id")

        return self.__pine.post({"resource": self.RESOURCE, "body": body})

    def revoke(self, invite_id: int) -> None:
        """
        Revoke an invite.

        Args:
            invite_id (int): organization invite id.

        Examples:
            >>> balena.models.organization.invite.revoke(2862)
        """

        self.__pine.delete({"resource": self.RESOURCE, "id": invite_id})

    def accept(self, invite_token: str) -> None:
        """
        Accepts an invite.

        Args:
            invite_token (str): invitation Token - invite token.

        """
        request(method="POST", settings=self.__settings, path=f"/org/v1/invitation/{invite_token}")


class OrganizationMembership:
    """
    This class implements organization membership model for balena python SDK.

    """

    def __init__(self, pine: PineClient, organization: Organization):
        self.__pine = pine
        self.__organization = organization
        self.tags = OrganizationMembershipTag(pine, organization)
        self.RESOURCE = "organization_membership"

    def get_all(self, options: AnyObject = {}) -> List[OrganizationMembershipType]:
        """
        Get all organization memberships.

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[OrganizationMembershipType]: organization memberships.

        Examples:
            >>> balena.models.organization.memberships.tags.get_all()
        """

        return self.__pine.get({"resource": self.RESOURCE, "options": options})

    def get_all_by_organization(
        self, handle_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[OrganizationMembershipType]:
        """
        Get all memberships by organization.

        Args:
            handle_or_id (Union[str, int]): organization handle (string) or id (number).
            options (AnyObject): extra pine options to use

        Returns:
            List[OrganizationMembershipType]: organization memberships.

        Examples:
            >>> balena.models.organization.memberships.get_all_by_organization(3014)
        """

        org_id = self.__organization.get(handle_or_id, {"$select": "id"})
        org_filter = {"$filter": {"is_member_of__organization": org_id}}
        return self.get_all(merge(org_filter, options))

    def get(self, membership_id: ResourceKey, options: AnyObject = {}) -> OrganizationMembershipType:
        """
        Get a single organization membership.

        Args:
            membership_id (ResourceKey): the id (int) or an object with the unique
            `user` & `is_member_of__organization` numeric pair of the membership
            options (AnyObject): extra pine options to use

        Returns:
            Organization membership.

        Examples:
            >>> balena.models.organization.memberships.get(17608)
        """
        result = self.__pine.get(
            {
                "resource": self.RESOURCE,
                "id": membership_id,  # type: ignore
                "options": options,
            }
        )

        if result is None:
            raise exceptions.OrganizationMembershipNotFound(membership_id)
        return result


class OrganizationMembershipTag(DependentResource[OrganizationMembershipTagType]):
    """
    This class implements organization membership tag model for balena python SDK.

    """

    def __init__(self, pine: PineClient, organization: Organization):
        self.__organization = organization
        super(OrganizationMembershipTag, self).__init__(
            "organization_membership_tag",
            "tag_key",
            "organization_membership",
            lambda id: self.__organization.get(id)["id"],
            pine,
        )

    def get_all_by_organization(
        self, handle_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[OrganizationMembershipTagType]:
        """
        Get all organization membership tags for an organization.

        Args:
            handle_or_id (Union[str, int]): organization handle (string) or id (number).
            options (AnyObject): extra pine options to use

        Returns:
            List[OrganizationMembershipTagType]: organization membership tags.

        Examples:
            >>> balena.models.organization.memberships.tags.get_all_by_organization(3014)
        """
        org_id = self.__organization.get(handle_or_id, {"$select": "id"})["id"]
        return super(OrganizationMembershipTag, self)._get_all(
            merge(
                {
                    "$filter": {
                        "organization_membership": {
                            "$any": {
                                "$alias": "om",
                                "$expr": {"om": {"is_member_of__organization": org_id}},
                            }
                        }
                    }
                },
                options,
            )
        )

    def get_all_by_organization_membership(
        self, membership_id: int, options: AnyObject = {}
    ) -> List[OrganizationMembershipTagType]:
        """
        Get all organization membership tags for a memberships of an organization.

        Args:
            membership_id (int): organization membership id.
            options (AnyObject): extra pine options to use

        Returns:
            list: organization membership tags.

        Examples:
            >>> balena.models.organization.memberships.tags.get_all_by_organization_membership(17608)
        """

        return super(OrganizationMembershipTag, self)._get_all_by_parent(membership_id, options)

    def get_all(self, options: AnyObject = {}) -> List[OrganizationMembershipTagType]:
        """
        Get all organization membership tags.

        Args:
            options (AnyObject): extra pine options to use


        Examples:
            >>> balena.models.organization.memberships.tags.get_all()
        """

        return super(OrganizationMembershipTag, self)._get_all(options)

    def get(self, membership_id: int, tag_key: str) -> Union[str, None]:
        """
        Get an organization membership tag.

        Args:
            membership_id: organization membership id.
            tag_key (str): tag key.

        Examples:
            >>> balena.models.organization.memberships.tags.get(17608, 'mTag1')
        """

        return super(OrganizationMembershipTag, self)._get(membership_id, tag_key)

    def set(self, membership_id: int, tag_key: str, value: str) -> None:
        """
        Set an organization membership tag.

        Args:
            membership_id: organization membership id.
            tag_key (str): tag key.
            value (str): tag value.

        Examples:
            >>> balena.models.organization.memberships.tags.set(17608, 'mTag1', 'Python SDK')
        """

        super(OrganizationMembershipTag, self)._set(membership_id, tag_key, value)

    def remove(self, membership_id: int, tag_key: str) -> None:
        """
        Remove an organization membership tag.

        Args:
            membership_id: organization membership id.
            tag_key (str): tag key.

        Examples:
            >>> balena.models.organization.memberships.tags.remove(17608, 'mTag1')
        """

        super(OrganizationMembershipTag, self)._remove(membership_id, tag_key)



from datetime import datetime, timedelta
from typing import List, Optional, Union

from .. import exceptions
from ..pine import PineClient
from ..types import AnyObject
from ..types.models import DeviceHistoryType
from ..utils import is_full_uuid, is_id, merge
from ..settings import Settings
from .application import Application


def history_timerange_filter_with_guard(from_date=None, to_date=None):
    from_date_filter = {}
    to_date_filter = {}

    if from_date is not None:
        if not isinstance(from_date, datetime):
            raise exceptions.InvalidParameter("from_date", from_date)
        else:
            from_date_filter = {"$ge": from_date}

    if to_date is not None:
        if not isinstance(to_date, datetime):
            raise exceptions.InvalidParameter("to_date", to_date)
        else:
            to_date_filter = {"$le": to_date}

    filter = {**from_date_filter, **to_date_filter}

    if filter == {}:
        return {}

    return {"created_at": filter}


class DeviceHistory:
    """
    This class implements device history model for balena python SDK.

    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__pine = pine
        self.__application = Application(pine, settings)

    def get_all_by_device(
        self,
        uuid_or_id: Union[str, int],
        from_date: datetime = datetime.utcnow() + timedelta(days=-7),
        to_date: Optional[datetime] = None,
        options: AnyObject = {},
    ) -> List[DeviceHistoryType]:
        """
        Get all device history entries for a device.

        Args:
            uuid_or_id (str): device uuid (32 / 62 digits string) or id (number) __note__: No short IDs supported
            from_date (datetime): history entries newer than or equal to this timestamp. Defaults to 7 days ago
            to_date (datetime): history entries younger or equal to this date.
            options (AnyObject): extra pine options to use

        Returns:
            List[DeviceHistoryType]: device history entries.

        Examples:
            >>> balena.models.device.history.get_all_by_device('6046335305c8142883a4466d30abe211')
            >>> balena.models.device.history.get_all_by_device(11196426)
            >>> balena.models.device.history.get_all_by_device(
            ...     11196426, from_date=datetime.utcnow() + timedelta(days=-5)
            ... )
            >>> balena.models.device.history.get_all_by_device(
            ...     11196426,
            ...     from_date=datetime.utcnow() + timedelta(days=-10),
            ...     to_date=from_date = datetime.utcnow() + timedelta(days=-5))
            ... )

        """
        dollar_filter = history_timerange_filter_with_guard(from_date, to_date)
        if is_id(uuid_or_id):
            dollar_filter = {**dollar_filter, "tracks__device": uuid_or_id}
        elif is_full_uuid(uuid_or_id):
            dollar_filter = {**dollar_filter, "uuid": uuid_or_id}
        else:
            raise exceptions.InvalidParameter("uuid_or_id", uuid_or_id)

        return self.__pine.get({"resource": "device_history", "options": merge({"$filter": dollar_filter}, options)})

    def get_all_by_application(
        self,
        slug_or_uuid_or_id: Union[str, int],
        from_date: datetime = datetime.utcnow() + timedelta(days=-7),
        to_date: Optional[datetime] = None,
        options: AnyObject = {},
    ) -> List[DeviceHistoryType]:
        """
        Get all device history entries for an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            from_date (datetime): history entries newer than or equal to this timestamp. Defaults to 7 days ago
            to_date (datetime): history entries younger or equal to this date.
            options (AnyObject): extra pine options to use

        Returns:
            List[DeviceHistoryType]: device history entries.

        Examples:
            >>> balena.models.device.history.get_all_by_application('myorg/myapp')
            >>> balena.models.device.history.get_all_by_application(11196426)
            >>> balena.models.device.history.get_all_by_application(
            ...     11196426, from_date=datetime.utcnow() + timedelta(days=-5)
            ... )
            >>> balena.models.device.history.get_all_by_application(
            ...     11196426,
            ...     from_date=datetime.utcnow() + timedelta(days=-10),
            ...     to_date=from_date = datetime.utcnow() + timedelta(days=-5))
            ... )
        """
        app_id = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})["id"]

        return self.__pine.get(
            {
                "resource": "device_history",
                "options": merge(
                    {
                        "$filter": {
                            **history_timerange_filter_with_guard(from_date, to_date),
                            "belongs_to__application": app_id,
                        }
                    },
                    options,
                ),
            }
        )



from typing import List

from .. import exceptions
from ..auth import Auth
from ..pine import PineClient
from ..types import AnyObject
from ..types.models import SSHKeyType
from ..settings import Settings


class Key:
    """
    This class implements ssh key model for balena python SDK.

    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__pine = pine
        self.__auth = Auth(pine, settings)

    def get_all(self, options: AnyObject = {}) -> List[SSHKeyType]:
        """
        Get all ssh keys.

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[SSHKeyType]: list of ssh keys.
        """
        return self.__pine.get({"resource": "user__has__public_key", "options": options})

    def get(self, id: int) -> SSHKeyType:
        """
        Get a single ssh key.

        Args:
            id (int): key id.

        Returns:
            SSHKeyType: ssh key info.
        """

        key = self.__pine.get({"resource": "user__has__public_key", "id": id})

        if key is None:
            raise exceptions.KeyNotFound(id)

        return key

    def remove(self, id: int) -> None:
        """
        Remove a ssh key.

        Args:
            id (int): key id.
        """

        self.__pine.delete({"resource": "user__has__public_key", "id": id})

    def create(self, title: str, key: str) -> SSHKeyType:
        """
        Create a ssh key.

        Args:
            title (str): key title.
            key (str): the public ssh key.

        Returns:
            SSHKeyType: new ssh key id.
        """
        # Avoid ugly whitespaces
        key = key.strip()

        user_id = self.__auth.get_user_info()["id"]
        return self.__pine.post(
            {"resource": "user__has__public_key", "body": {"title": title, "public_key": key, "user": user_id}}
        )



import configparser
import os
import os.path as Path
import shutil
import sys
from typing import Dict, TypedDict, Optional, Union, Literal
from abc import ABC, abstractmethod
from copy import deepcopy

from . import exceptions
from .resources import Message


class SettingsConfig(TypedDict, total=False):
    balena_host: str
    api_version: str
    device_actions_endpoint_version: str
    data_directory: Union[str, Literal[False]]
    image_cache_time: str
    token_refresh_interval: str
    timeout: str
    request_limit: str
    request_limit_interval: str
    retry_rate_limited_request: bool


class SettingsProviderInterface(ABC):
    @abstractmethod
    def has(self, key: str) -> bool:
        pass

    @abstractmethod
    def get(self, key: str) -> Union[str, bool]:
        pass

    @abstractmethod
    def get_all(self) -> Dict[str, Union[str, bool]]:
        pass

    @abstractmethod
    def set(self, key: str, value: Union[str, bool]) -> None:
        pass

    @abstractmethod
    def remove(self, key: str) -> bool:
        pass


DEFAULT_SETTINGS = {
    # These are default config values
    "balena_host": "balena-cloud.com",
    "api_version": "v6",
    "device_actions_endpoint_version": "v1",
    # cache time : 1 week in milliseconds
    "image_cache_time": str(1 * 1000 * 60 * 60 * 24 * 7),
    # token refresh interval: 1 hour in milliseconds
    "token_refresh_interval": str(1 * 1000 * 60 * 60),
    # requests timeout: 30 seconds in milliseconds
    "timeout": str(30 * 1000),
    # requests timeout: 60 seconds in seconds
    "request_limit_interval": str(60),
    "retry_rate_limited_request": False,
}


class FileStorageSettingsProvider(SettingsProviderInterface):
    """
    This class handles settings for balena python SDK.

    Attributes:
        HOME_DIRECTORY (str): home directory path.
        CONFIG_SECTION (str): section name in configuration file.
        CONFIG_FILENAME (str): configuration file name.
        _setting (dict): default value to settings.

    """

    HOME_DIRECTORY = os.getenv("BALENA_SETTINGS_HOME_DIRECTORY", default=Path.expanduser("~"))
    CONFIG_SECTION = os.getenv("BALENA_SETTINGS_CONFIG_SECTION", default="Settings")
    CONFIG_FILENAME = os.getenv("BALENA_SETTINGS_CONFIG_FILENAME", default="balena.cfg")
    DEFAULT_SETTING_KEYS = set(
        [
            "builder_url",
            "pine_endpoint",
            "api_endpoint",
            "api_version",
            "data_directory",
            "image_cache_time",
            "token_refresh_interval",
            "cache_directory",
            "timeout",
            "device_actions_endpoint_version",
            "retry_rate_limited_request",
        ]
    )

    def __init__(self, settings_config: Optional[SettingsConfig]):
        _base_settings = deepcopy(DEFAULT_SETTINGS)

        if settings_config is not None:
            _base_settings = {**_base_settings, **settings_config}

        host = _base_settings["balena_host"]
        _base_settings["builder_url"] = f"https://builder.{host}/"
        _base_settings["api_endpoint"] = f"https://api.{host}/"
        _base_settings["pine_endpoint"] = f"https://api.{host}/{_base_settings['api_version']}/"

        data_directory = _base_settings.get("data_directory")
        if data_directory is None or data_directory is True:
            _base_settings["data_directory"] = Path.join(FileStorageSettingsProvider.HOME_DIRECTORY, ".balena")

        _base_settings["cache_directory"] = Path.join(_base_settings["data_directory"], "cache")

        self.__base_settings = _base_settings
        self._setting = _base_settings

        config_file_path = Path.join(self._setting["data_directory"], self.CONFIG_FILENAME)
        try:
            self.__read_settings()
            if not self.DEFAULT_SETTING_KEYS.issubset(set(self._setting)):
                raise
        except Exception:
            # Backup old settings file if it exists.
            try:
                if Path.isfile(config_file_path):
                    shutil.move(
                        config_file_path,
                        Path.join(
                            self._setting["data_directory"],
                            "{0}.{1}".format(self.CONFIG_FILENAME, "old"),
                        ),
                    )
            except OSError:
                pass
            self.__write_settings(default=True)
            print(Message.INVALID_SETTINGS.format(path=config_file_path), file=sys.stderr)

    def __write_settings(self, default=None):
        """
        Write settings to file.

        Args:
            default (Optional[bool]): write default settings.

        """

        if default:
            self._setting = self.__base_settings
        config = configparser.ConfigParser()
        config.add_section(self.CONFIG_SECTION)
        for key in self._setting:
            value = self._setting[key]
            if isinstance(value, bool):
                value = "true" if value else "false"
            config.set(self.CONFIG_SECTION, key, value)
        if not Path.isdir(self._setting["data_directory"]):
            os.makedirs(self._setting["data_directory"])
        with open(Path.join(self._setting["data_directory"], self.CONFIG_FILENAME), "w") as config_file:
            config.write(config_file)

    def __read_settings(self):
        config_reader = configparser.ConfigParser()
        config_reader.read(Path.join(self._setting["data_directory"], self.CONFIG_FILENAME))
        config_data = {}
        options = config_reader.options(self.CONFIG_SECTION)
        for option in options:
            try:
                config_data[option] = config_reader.get(self.CONFIG_SECTION, option)
                if config_data[option] == "true":
                    config_data[option] = True
                if config_data[option] == "false":
                    config_data[option] = False
            except Exception:
                config_data[option] = None
        self._setting = config_data

    def has(self, key: str) -> bool:
        self.__read_settings()
        if key in self._setting:
            return True
        return False

    def get(self, key: str) -> Union[str, bool]:
        try:
            self.__read_settings()
            return self._setting[key]
        except KeyError:
            raise exceptions.InvalidOption(key)

    def get_all(self) -> Dict[str, Union[str, bool]]:
        self.__read_settings()
        return self._setting

    def set(self, key: str, value: Union[str, bool]) -> None:
        self._setting[key] = str(value)
        self.__write_settings()

    def remove(self, key: str) -> bool:
        # if key is not in settings, return False
        result = self._setting.pop(key, False)
        if result is not False:
            self.__write_settings()
            return True
        return False


class InMemorySettingsProvider(SettingsProviderInterface):
    def __init__(self, settings_config: Optional[SettingsConfig]):
        self._settings = deepcopy(DEFAULT_SETTINGS)

        if settings_config is not None:
            self._settings = {**DEFAULT_SETTINGS, **settings_config}

        host = self._settings["balena_host"]
        self._settings["builder_url"] = f"https://builder.{host}/"
        self._settings["api_endpoint"] = f"https://api.{host}/"
        self._settings["pine_endpoint"] = f"https://api.{host}/{self._settings['api_version']}/"

    def has(self, key: str) -> bool:
        if key in self._settings:
            return True
        return False

    def get(self, key: str) -> Union[str, bool]:
        try:
            return self._settings[key]
        except KeyError:
            raise exceptions.InvalidOption(key)

    def get_all(self) -> Dict[str, Union[str, bool]]:
        return self._settings

    def set(self, key: str, value: Union[str, bool]) -> None:
        self._settings[key] = str(value)

    def remove(self, key: str) -> bool:
        result = self._settings.pop(key, False)
        if result:
            return True
        return False


class Settings(SettingsProviderInterface):
    """
    This class handles settings for balena python SDK.

    """

    def __init__(self, settings_config: Optional[SettingsConfig]):
        self.__settings_provider: SettingsProviderInterface

        if settings_config and settings_config.get("data_directory") is False:
            self.__settings_provider = InMemorySettingsProvider(settings_config)
        else:
            self.__settings_provider = FileStorageSettingsProvider(settings_config)

    def has(self, key: str) -> bool:
        """
        Check if a setting exists.

        Args:
            key (str): setting.

        Returns:
            bool: True if exists, False otherwise.

        Examples:
            >>> balena.settings.has('api_endpoint')
        """
        return self.__settings_provider.has(key)

    def get(self, key: str) -> Union[str, bool]:
        """
        Get a setting value.

        Args:
            key (str): setting.

        Returns:
            str: setting value.

        Raises:
            InvalidOption: If getting a non-existent setting.

        Examples:
            >>> balena.settings.get('api_endpoint')
        """
        return self.__settings_provider.get(key)

    def get_all(self) -> Dict[str, Union[str, bool]]:
        """
        Get all settings.

        Returns:
            dict: all settings.

        Examples:
            >>> balena.settings.get_all()
        """
        return self.__settings_provider.get_all()

    def set(self, key: str, value: Union[str, bool]) -> None:
        """
        Set value for a setting.

        Args:
            key (str): setting.
            value (str): setting value.

        Examples:
            >>> balena.settings.set(key='tmp',value='123456')
        """
        return self.__settings_provider.set(key, value)

    def remove(self, key: str) -> bool:
        """
        Remove a setting.

        Args:
            key (str): setting.

        Returns:
            bool: True if successful, False otherwise.

        Examples:
            # Remove an existing key from settings
            >>> balena.settings.remove('tmp')
            # Remove a non-existing key from settings
            >>> balena.settings.remove('tmp1')
        """
        return self.__settings_provider.remove(key)



from datetime import datetime
from math import isinf
from typing import List, Literal, Optional, Union, cast
from urllib.parse import urljoin

from .. import exceptions
from ..balena_auth import request
from ..dependent_resource import DependentResource
from ..pine import PineClient
from ..settings import Settings
from ..types import AnyObject, ApplicationInviteOptions, ApplicationMembershipRoles, ResourceKey, ShutdownOptions
from ..types.models import (
    ApplicationInviteType,
    ApplicationMembershipType,
    TypeApplication,
    BaseTagType,
    EnvironmentVariableBase,
    TypeApplicationWithDeviceServiceDetails,
)
from ..utils import (
    generate_current_service_details,
    get_current_service_details_pine_expand,
    is_id,
    merge,
    with_supervisor_locked_error,
)
from .device_type import DeviceType


class Application:
    """
    This class implements application model for balena python SDK.
    """

    def __init__(self, pine: PineClient, settings: Settings, load_inner_models=True):
        if load_inner_models:
            self.__device_type = DeviceType(pine, settings)
            self.__release = Release(pine, settings)

        self.__pine = pine
        self.__settings = settings
        self.tags = ApplicationTag(pine, self)
        self.config_var = ApplicationConfigVariable(pine, self)
        self.env_var = ApplicationEnvVariable(pine, self)
        self.build_var = BuildEnvVariable(pine, self)
        self.membership = ApplicationMembership(pine, self)
        self.invite = ApplicationInvite(pine, self, settings)

    def __get_access_filter(self):
        return {
            "is_directly_accessible_by__user": {
                "$any": {
                    "$alias": "dau",
                    "$expr": {
                        "1": 1,
                    },
                },
            },
        }

    def __get_device_type_id(self, device_type: str) -> int:
        dt = self.__device_type.get(
            device_type,
            {
                "$select": "id",
                "$expand": {
                    "is_default_for__application": {
                        "$select": "is_archived",
                        "$filter": {
                            "is_host": True,
                        },
                    },
                },
            },
        )

        host_apps = dt.get("is_default_for__application", [])
        if len(host_apps) > 0 and all(map(lambda ha: ha["is_archived"], host_apps)):
            raise exceptions.BalenaDiscontinuedDeviceType(device_type)

        return dt["id"]

    def __get_organization_id(self, organization: Union[str, int]) -> int:
        id_filter = {"handle": organization}
        if is_id(organization):
            id_filter = {"id": organization}

        org = self.__pine.get(
            {
                "resource": "organization",
                "id": id_filter,
                "options": {"$select": ["id"]},
            }
        )

        if org is None:
            raise exceptions.OrganizationNotFound(organization)

        return org["id"]

    def get_id(self, slug_or_uuid_or_id: Union[str, int]) -> int:
        """
        Given an application slug or uuid or id, returns it numeric id.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)

        Returns:
            int: The id.

        Examples:
            >>> balena.models.application.get_dashboard_url(1476418)

        """
        if is_id(slug_or_uuid_or_id):
            return int(slug_or_uuid_or_id)
        app = self.get(slug_or_uuid_or_id, {"$select": "id"})
        return app["id"]

    def get_dashboard_url(self, app_id: int) -> str:
        """
        Get Dashboard URL for a specific application.

        Args:
            app_id (int): application id.

        Returns:
            str: Dashboard URL for the specific application.

        Examples:
            >>> balena.models.application.get_dashboard_url(1476418)
        """
        try:
            if isinf(int(app_id)):
                raise exceptions.InvalidParameter("app_id", app_id)
        except ValueError:
            raise exceptions.InvalidParameter("app_id", app_id)

        return urljoin(
            cast(str, self.__settings.get("api_endpoint")).replace("api", "dashboard"),
            f"/apps/{app_id}",
        )

    def get_all(
        self,
        options: AnyObject = {},
        context: Optional[str] = "directly_accessible",
    ) -> List[TypeApplication]:
        """
        Get all applications

        Args:
            options (AnyObject): extra pine options to use
            context (Optional[str]): extra access filters, None or 'directly_accessible'

        Returns:
            List[APIKeyType]: user API key

        Examples:
            >>> balena.models.application.get_all()
        """

        return self.__pine.get(
            {
                "resource": "application",
                "options": merge(
                    {
                        **({"$filter": self.__get_access_filter()} if context == "directly_accessible" else {}),
                        "$orderby": "app_name asc",
                    },
                    options,
                ),
            }
        )

    def get_all_directly_accessible(
        self,
        options: AnyObject = {},
    ) -> List[TypeApplication]:
        """
        Get all applications directly accessible by the user

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[APIKeyType]: user API key

        Examples:
            >>> balena.models.application.get_all_directly_accessible()
        """

        return self.get_all(options, "directly_accessible")

    def get(
        self,
        slug_or_uuid_or_id: Union[str, int],
        options: AnyObject = {},
        context: Optional[str] = None,
    ) -> TypeApplication:
        """
        Get a single application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use
            context (Optional[str]): extra access filters, None or 'directly_accessible'

        Returns:
            TypeApplication: application info.

        Examples:
            >>> balena.models.application.get("myorganization/myapp")
            >>> balena.models.application.get(123)

        """

        access_filter = None
        if context == "directly_accessible":
            access_filter = self.__get_access_filter()

        application = None
        if is_id(slug_or_uuid_or_id):
            application = self.__pine.get(
                {
                    "resource": "application",
                    "id": slug_or_uuid_or_id,
                    "options": merge(
                        {} if access_filter is None else {"$filter": access_filter},
                        options,
                    ),
                }
            )
        elif isinstance(slug_or_uuid_or_id, str):
            lower_case_slug_or_uuid = slug_or_uuid_or_id.lower()

            if access_filter is not None:
                app_filter = {
                    **access_filter,
                    "$or": {
                        "slug": lower_case_slug_or_uuid,
                        "uuid": lower_case_slug_or_uuid,
                    },
                }
            else:
                app_filter = {
                    "$or": {
                        "slug": lower_case_slug_or_uuid,
                        "uuid": lower_case_slug_or_uuid,
                    }
                }
            applications = self.__pine.get(
                {
                    "resource": "application",
                    "options": merge({"$filter": app_filter}, options),
                }
            )

            if len(applications) > 1:
                raise exceptions.AmbiguousApplication(slug_or_uuid_or_id)
            try:
                application = applications[0]
            except IndexError:
                raise exceptions.ApplicationNotFound(slug_or_uuid_or_id)

        if application is None:
            raise exceptions.ApplicationNotFound(slug_or_uuid_or_id)

        return application

    def get_directly_accessible(
        self,
        slug_or_uuid_or_id: Union[str, int],
        options: AnyObject = {},
    ) -> TypeApplication:
        """
        Get a single application directly accessible by the user

        Args:
            slug_or_uuid_or_id (str): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            TypeApplication: application info.

        Examples:
            >>> balena.models.application.get_directly_accessible("myorganization/myapp")
            >>> balena.models.application.get_directly_accessible(123)
        """
        return self.get(slug_or_uuid_or_id, options, "directly_accessible")

    def get_with_device_service_details(
        self,
        slug_or_uuid_or_id: Union[str, int],
        options: AnyObject = {},
    ) -> TypeApplicationWithDeviceServiceDetails:
        """
        This method does not map exactly to the underlying model: it runs a
        larger prebuilt query, and reformats it into an easy to use and
        understand format. If you want more control, or to see the raw model
        directly, use `application.get(uuidOrId, options)` instead.

        Args:
            slug_or_uuid_or_id (str): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            TypeApplication: application info.

        Examples:
            >>> balena.models.application.get_with_device_service_details('my_org_handle/my_app_name')
        """
        service_options = merge(
            {"$expand": [{"owns__device": {"$expand": get_current_service_details_pine_expand(True)}}]},
            options,
        )

        app = self.get(slug_or_uuid_or_id, service_options)

        devices = app.get("owns__device")
        if app is not None and devices is not None:
            app["owns__device"] = list(map(generate_current_service_details, devices))

        return app

    def get_by_name(
        self,
        app_name: str,
        options: AnyObject = {},
        context: Optional[str] = "directly_accessible",
    ) -> TypeApplication:
        """
         Get a single application using the appname.

        Args:
            slug_or_uuid_or_id (str): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use
            context (Optional[str]): extra access filters, None or 'directly_accessible'

        Returns:
            TypeApplication: application info.

        Examples:
            >>> balena.models.application.get("myapp")
        """
        apps = self.__pine.get(
            {
                "resource": "application",
                "options": merge(
                    {
                        "$filter": {
                            **(self.__get_access_filter() if context == "directly_accessible" else {}),
                            "app_name": app_name,
                        }
                    },
                    options,
                ),
            }
        )

        if len(apps) == 0:
            raise exceptions.ApplicationNotFound(app_name)

        if len(apps) > 1:
            raise exceptions.AmbiguousApplication(app_name)

        return apps[0]

    def get_by_owner(self, app_name: str, owner: str, options: AnyObject = {}) -> TypeApplication:
        """
        Get a single application using the appname and the handle of the owning organization.

        Args:
            app_name (str): application name.
            owner (str): The handle of the owning organization.
            options (AnyObject): extra pine options to use.

        Returns:
            TypeApplication: application info.

        Examples:
            >>> balena.models.application.get_by_owner('foo', 'my_org')
        """

        slug = f"{owner.lower()}/{app_name.lower()}"
        app = self.__pine.get(
            {
                "resource": "application",
                "id": {"slug": slug},
                "options": options,
            }
        )

        if app is None:
            raise exceptions.ApplicationNotFound(slug)

        return app

    def has(self, slug_or_uuid_or_id: Union[str, int]) -> bool:
        """
        Check if an application exists.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)

        Returns:
            bool: True if application exists, False otherwise.

        Examples:
            >>> balena.models.application.has('my_org/foo')
        """

        try:
            self.get(slug_or_uuid_or_id, {"$select": ["id"]})
            return True
        except exceptions.ApplicationNotFound:
            return False

    def has_any(self) -> bool:
        """
        Check if the user has any applications.

        Returns:
            bool: True if user has any applications, False otherwise.

        Examples:
            >>> balena.models.application.has_any()
        """

        applications = self.get_all(
            {"$select": ["id"]},
            "directly_accessible",
        )
        return len(applications) != 0

    def create(
        self,
        name: str,
        device_type: str,
        organization: Union[str, int],
        application_class: Optional[Literal["app", "fleet", "block"]] = None,
    ) -> TypeApplication:
        """
        Create an application.

        Args:
            name (str): application name.
            device_type (str): device type (slug).
            organization (Union[str, int]): handle or id of the organization that the application will belong to.
            application_class (Optional[Literal["app", "fleet", "block"]]): application class.

        Returns:
            TypeApplication: application info.

        Examples:
            >>> balena.models.application.create('foo', 'raspberry-pi', 12345)
            >>> balena.models.application.create('foo', 'raspberry-pi', 12345, 'block')
        """

        if organization is None:
            raise exceptions.InvalidParameter("organization", organization)

        # TODO: run these two in parallel
        device_type_id = self.__get_device_type_id(device_type)
        organization_id = self.__get_organization_id(organization)

        body = {
            "app_name": name,
            "is_for__device_type": device_type_id,
            "organization": organization_id,
        }

        if application_class is not None:
            body["is_of__class"] = application_class

        return self.__pine.post({"resource": "application", "body": body})

    # TODO: enable batch operations
    def remove(self, slug_or_uuid_or_id: Union[str, int]) -> None:
        """
        Remove application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).

        Examples:
            >>> balena.models.application.remove('my_org/my_app')
            >>> balena.models.application.remove('c184556293854781aea71b0bdae10e45')
            >>> balena.models.application.remove(123)
        """

        try:
            application_id = self.get_id(slug_or_uuid_or_id)
            self.__pine.delete({"resource": "application", "id": application_id})
        except exceptions.RequestError as e:
            if e.status_code == 404:
                raise exceptions.ApplicationNotFound(slug_or_uuid_or_id)
            raise e

    def rename(self, slug_or_uuid_or_id: Union[str, int], new_name: str) -> None:
        """
        Rename application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            new_name (str): new application name.

        Examples:
            >>> balena.models.application.rename(1681618, 'py-test-app')
        """

        try:
            application_id = self.get_id(slug_or_uuid_or_id)
            self.__pine.patch(
                {
                    "resource": "application",
                    "id": application_id,
                    "body": {"app_name": new_name},
                }
            )
        except exceptions.RequestError as e:
            if e.status_code == 404:
                raise exceptions.ApplicationNotFound(slug_or_uuid_or_id)
            raise e

    def restart(self, slug_or_uuid_or_id: Union[str, int]):
        """
        Restart application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).

        Examples:
            >>> balena.models.application.restart('myorg/RPI1')
        """

        def __restart():
            try:
                application_id = self.get_id(slug_or_uuid_or_id)
                request(method="POST", path=f"/applcation/{application_id}/restart", settings=self.__settings)
            except exceptions.RequestError as e:
                if e.status_code == 404:
                    raise exceptions.ApplicationNotFound(slug_or_uuid_or_id)
                raise e

        with_supervisor_locked_error(__restart)

    def generate_provisioning_key(
        self,
        slug_or_uuid_or_id: Union[str, int],
        key_name: Optional[str] = None,
        description: Optional[str] = None,
        expiry_date: Optional[str] = None,
    ) -> str:
        """
        Generate a device provisioning key for a specific application.

        Args:
            slug_or_uuid_or_id (str): application slug (string), uuid (string) or id (number)
            key_name (Optional[str]): provisioning key name.
            description (Optional[str]): description for provisioning key.
            expiry_date (Optional[str]): expiry date for provisioning key, for example: `2030-01-01T00:00:00Z`.

        Returns:
            str: device provisioning key.

        Examples:
            >>> balena.models.application.generate_provisioning_key(5685)
        """
        try:
            application_id = self.get_id(slug_or_uuid_or_id)
            return request(
                method="POST",
                path="/api-key/v1/",
                settings=self.__settings,
                body={
                    "actorType": "application",
                    "actorTypeId": application_id,
                    "roles": ["provisioning-api-key"],
                    "name": key_name,
                    "description": description,
                    "expiryDate": expiry_date,
                },
            ).strip('"')
        except exceptions.RequestError as e:
            if e.status_code == 404:
                raise exceptions.ApplicationNotFound(slug_or_uuid_or_id)
            raise e

    def purge(self, app_id: int) -> None:
        """
        Purge devices by application id

        Args:
            app_id (int): application id (number)

        Examples:
            >>> balena.models.application.purge(5685)
        """
        with_supervisor_locked_error(
            lambda: request(
                method="POST",
                path="/supervisor/v1/purge",
                settings=self.__settings,
                body={"appId": app_id, "data": {"appId": f"{app_id}"}},
            )
        )

    def shutdown(self, app_id: int, options: ShutdownOptions = {}) -> None:
        """
        Shutdown devices by application id

        Args:
            app_id (int): application id (number)
            options (ShutdownOptions): override update lock

        Examples:
            >>> balena.models.application.shutdown(5685)
            >>> balena.models.application.shutdown(5685, {"force": True})
        """
        with_supervisor_locked_error(
            lambda: request(
                method="POST",
                path="/supervisor/v1/shutdown",
                settings=self.__settings,
                body={
                    "appId": app_id,
                    "data": {"force": bool(options.get("force"))},
                },
            )
        )

    def reboot(self, app_id: int, options: ShutdownOptions = {}) -> None:
        """
        Reboots devices by application id

        Args:
            app_id (int): application id (number)
            options (ShutdownOptions): override update lock

        Examples:
            >>> balena.models.application.reboot(5685)
            >>> balena.models.application.reboot(5685, {"force": True})
        """
        with_supervisor_locked_error(
            lambda: request(
                method="POST",
                path="/supervisor/v1/reboot",
                settings=self.__settings,
                body={
                    "appId": app_id,
                    "data": {"force": bool(options.get("force"))},
                },
            )
        )

    def will_track_new_releases(self, slug_or_uuid_or_id: Union[str, int]) -> bool:
        """
         Get whether the application is configured to receive updates whenever a new release is available.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).

        Returns:
            bool: is tracking the latest release.

        Examples:
            >>> balena.models.application.will_track_new_releases(5685)
        """

        app = self.get(slug_or_uuid_or_id, {"$select": "should_track_latest_release"})
        return bool(app.get("should_track_latest_release"))

    def is_tracking_latest_release(self, slug_or_uuid_or_id: Union[str, int]) -> bool:
        """
        Get whether the application is up to date and is tracking the latest finalized release for updates

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).

        Returns:
            bool: is tracking the latest release.

        Examples:
            >>> balena.models.application.is_tracking_latest_release(5685)
        """
        app_options = {
            "$select": "should_track_latest_release",
            "$expand": {
                "should_be_running__release": {"$select": "id"},
                "owns__release": {
                    "$select": "id",
                    "$top": 1,
                    "$filter": {
                        "is_final": True,
                        "is_passing_tests": True,
                        "is_invalidated": False,
                        "status": "success",
                    },
                    "$orderby": "created_at desc",
                },
            },
        }

        app = self.get(slug_or_uuid_or_id, app_options)
        tracked_release = app["should_be_running__release"][0]
        latest_release = app["owns__release"][0]

        return bool(
            app.get("should_track_latest_release")
            and (not latest_release or tracked_release.get("id") == latest_release["id"])
        )

    def pin_to_release(self, slug_or_uuid_or_id: Union[str, int], full_release_hash: str) -> None:
        """
        Configures the application to run a particular release
        and not get updated when the latest release changes.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            full_release_hash (str) : the hash of a successful release (string)

        Examples:
            >>> balena.models.application.set_to_release(5685, '7dba4e0c461215374edad74a5b78f470b894b5b7')
        """

        application_id = self.get_id(slug_or_uuid_or_id)
        release = self.__release.get(
            full_release_hash,
            {
                "$select": "id",
                "$top": 1,
                "$filter": {
                    "belongs_to__application": application_id,
                    "status": "success",
                },
            },
        )

        self.__pine.patch(
            {
                "resource": "application",
                "id": application_id,
                "body": {
                    "should_be_running__release": release["id"],
                    "should_track_latest_release": False,
                },
            }
        )

    def get_target_release_hash(self, slug_or_uuid_or_id: Union[str, int]) -> Optional[str]:
        """
        Get the hash of the current release for a specific application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)

        Returns:
            Optional[str]: The release hash of the current release or None.

        Examples:
            >>> balena.models.application.get_target_release_hash(5685)
        """
        app_options = {
            "$select": "id",
            "$expand": {"should_be_running__release": {"$select": "commit"}},
        }
        application = self.get(slug_or_uuid_or_id, app_options)

        return application.get("should_be_running__release", [{}])[0].get("commit")

    def track_latest_release(self, slug_or_uuid_or_id: Union[str, int]) -> None:
        """
        Configure a specific application to track the latest available release.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)

        Examples:
            >>> balena.models.application.track_latest_release(5685)
        """

        app_options = {
            "$select": "id",
            "$expand": {
                "owns__release": {
                    "$select": "id",
                    "$top": 1,
                    "$filter": {
                        "is_final": True,
                        "is_passing_tests": True,
                        "is_invalidated": False,
                        "status": "success",
                    },
                    "$orderby": "created_at desc",
                }
            },
        }

        application = self.get(slug_or_uuid_or_id, app_options)
        body = {"should_track_latest_release": True}
        latest_release = application.get("owns__release", [None])[0]
        if latest_release is not None:
            body["should_be_running__release"] = latest_release.get("id")

        self.__pine.patch(
            {
                "resource": "application",
                "id": application["id"],
                "body": body,
            }
        )

    def enable_device_urls(self, slug_or_uuid_or_id: Union[str, int]) -> None:
        """
        Enable device urls for all devices that belong to an application

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).

        Examples:
            >>> balena.models.application.enable_device_urls(5685)
        """

        app = self.get(slug_or_uuid_or_id, {"$select": "id"})
        self.__pine.patch(
            {
                "resource": "device",
                "body": {"is_web_accessible": True},
                "options": {"$filter": {"belongs_to__application": app["id"]}},
            }
        )

    def disable_device_urls(self, slug_or_uuid_or_id: Union[str, int]) -> None:
        """
        Disable device urls for all devices that belong to an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).

        Examples:
            >>> balena.models.application.disable_device_urls(5685)
        """

        app = self.get(slug_or_uuid_or_id, {"$select": "id"})
        self.__pine.patch(
            {
                "resource": "device",
                "body": {"is_web_accessible": False},
                "options": {"$filter": {"belongs_to__application": app["id"]}},
            }
        )

    def grant_support_access(self, slug_or_uuid_or_id: Union[str, int], expiry_timestamp: int):
        """
        Grant support access to an application until a specified time.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            expiry_timestamp (int): a timestamp in ms for when the support access will expire.

        Examples:
            >>> balena.models.application.grant_support_access(5685, 1511974999000)
        """

        if expiry_timestamp is None or expiry_timestamp <= int(
            (datetime.utcnow() - datetime.utcfromtimestamp(0)).total_seconds() * 1000
        ):
            raise exceptions.InvalidParameter("expiry_timestamp", expiry_timestamp)

        try:
            application_id = self.get_id(slug_or_uuid_or_id)
            self.__pine.patch(
                {
                    "resource": "application",
                    "id": application_id,
                    "body": {"is_accessible_by_support_until__date": expiry_timestamp},
                }
            )
        except exceptions.RequestError as e:
            if e.status_code == 404:
                raise exceptions.ApplicationNotFound(slug_or_uuid_or_id)
            raise e

    def revoke_support_access(self, slug_or_uuid_or_id: Union[str, int]):
        """
        Revoke support access to an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).

        Examples:
            >>> balena.models.application.revoke_support_access(5685)
        """

        try:
            application_id = self.get_id(slug_or_uuid_or_id)
            self.__pine.patch(
                {
                    "resource": "application",
                    "id": application_id,
                    "body": {"is_accessible_by_support_until__date": None},
                }
            )
        except exceptions.RequestError as e:
            if e.status_code == 404:
                raise exceptions.ApplicationNotFound(slug_or_uuid_or_id)
            raise e


class ApplicationTag(DependentResource[BaseTagType]):
    """
    This class implements application tag model for balena python SDK.

    """

    def __init__(self, pine: PineClient, application: Application):
        self.__application = application
        super(ApplicationTag, self).__init__(
            "application_tag",
            "tag_key",
            "application",
            lambda id: self.__application.get(id, {"$select": "id"})["id"],
            pine,
        )

    def get_all_by_application(self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[BaseTagType]:
        """
        Get all application tags for an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[BaseTagType]: tags list.

        Examples:
            >>> balena.models.application.tags.get_all_by_application(1005160)
        """
        return super(ApplicationTag, self)._get_all_by_parent(slug_or_uuid_or_id, options)

    def set(self, slug_or_uuid_or_id: Union[str, int], tag_key: str, value: str) -> None:
        """
        Set an application tag (update tag value if it exists).

        Args:
            slug_or_uuid_or_id (int): application slug (string), uuid (string) or id (number)
            tag_key (str): tag key.
            value (str): tag value.

        Returns:
            List[BaseTagType]: tags list.

        Examples:
            >>> balena.models.application.tags.set(1005767, 'tag1', 'Python SDK')
        """
        super(ApplicationTag, self)._set(slug_or_uuid_or_id, tag_key, value)

    def remove(self, slug_or_uuid_or_id: Union[str, int], tag_key: str) -> None:
        """
        Remove an application tag.

        Args:
            slug_or_uuid_or_id (int): application slug (string), uuid (string) or id (number)
            tag_key (str): tag key.

        Examples:
            >>> balena.models.application.tags.remove(1005767, 'tag1')
        """
        super(ApplicationTag, self)._remove(slug_or_uuid_or_id, tag_key)


class ApplicationConfigVariable(DependentResource[EnvironmentVariableBase]):
    """
    This class implements application config variable model for balena python SDK.

    """

    def __init__(self, pine: PineClient, application: Application):
        self.__application = application
        super(ApplicationConfigVariable, self).__init__(
            "application_config_variable",
            "name",
            "application",
            lambda id: self.__application.get(id, {"$select": "id"})["id"],
            pine,
        )

    def get_all_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[EnvironmentVariableBase]:
        """
        Get all application config variables by application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: application config variables.

        Examples:
            >>> balena.models.application.config_var.get_all_by_application(9020)
        """
        return super(ApplicationConfigVariable, self)._get_all_by_parent(slug_or_uuid_or_id, options)

    def get(self, slug_or_uuid_or_id: Union[str, int], env_var_name: str) -> Optional[str]:
        """
        Get application config variable.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            env_var_name (str): environment variable name.

        Examples:
            >>> balena.models.application.config_var.get('8deb12','test_env4')
        """
        return super(ApplicationConfigVariable, self)._get(slug_or_uuid_or_id, env_var_name)

    def set(self, slug_or_uuid_or_id: Union[str, int], env_var_name: str, value: str) -> None:
        """
        Set the value of a specific application config variable.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            env_var_name (str): environment variable name.
            value (str): environment variable value.

        Examples:
            >>> balena.models.application.config_var.set('8deb12','test_env', 'testing1')
        """
        super(ApplicationConfigVariable, self)._set(slug_or_uuid_or_id, env_var_name, value)

    def remove(self, slug_or_uuid_or_id: Union[str, int], key: str) -> None:
        """
        Remove an application config variable.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)

        Examples:
            >>> balena.models.application.config_var.remove(2184)
        """
        super(ApplicationConfigVariable, self)._remove(slug_or_uuid_or_id, key)


class ApplicationEnvVariable(DependentResource[EnvironmentVariableBase]):
    """
    This class implements application environment variable model for balena python SDK.

    """

    def __init__(self, pine: PineClient, application: Application):
        self.__application = application
        super(ApplicationEnvVariable, self).__init__(
            "application_environment_variable",
            "name",
            "application",
            lambda id: self.__application.get(id, {"$select": "id"})["id"],
            pine,
        )

    def get_all_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[EnvironmentVariableBase]:
        """
        Get all application environment variables by application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: application environment variables.

        Examples:
            >>> balena.models.application.env_var.get_all_by_application(9020)
            >>> balena.models.application.env_var.get_all_by_application("myorg/myslug")
        """
        return super(ApplicationEnvVariable, self)._get_all_by_parent(slug_or_uuid_or_id, options)

    def get(self, slug_or_uuid_or_id: Union[str, int], env_var_name: str) -> Optional[str]:
        """
        Get application environment variable.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            env_var_name (str): environment variable name.

        Examples:
            >>> balena.models.application.env_var.get('8deb12','test_env4')
        """
        return super(ApplicationEnvVariable, self)._get(slug_or_uuid_or_id, env_var_name)

    def set(self, slug_or_uuid_or_id: Union[str, int], env_var_name: str, value: str) -> None:
        """
        Set the value of a specific application environment variable.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            env_var_name (str): environment variable name.
            value (str): environment variable value.

        Examples:
            >>> balena.models.application.env_var.set('8deb12','test_env4', 'testing1')
        """
        super(ApplicationEnvVariable, self)._set(slug_or_uuid_or_id, env_var_name, value)

    def remove(self, slug_or_uuid_or_id: Union[str, int], key: str) -> None:
        """
        Remove an application environment variable.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)

        Examples:
            >>> balena.models.application.env_var.remove(2184)
        """
        super(ApplicationEnvVariable, self)._remove(slug_or_uuid_or_id, key)


class BuildEnvVariable(DependentResource[EnvironmentVariableBase]):
    """
    This class implements build environment variable model for balena python SDK.

    """

    def __init__(self, pine: PineClient, application: Application):
        self.__application = application
        super(BuildEnvVariable, self).__init__(
            "build_environment_variable",
            "name",
            "application",
            lambda id: self.__application.get(id, {"$select": "id"})["id"],
            pine,
        )

    def get_all_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[EnvironmentVariableBase]:
        """
        Get all build environment variables by application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: build environment variables.

        Examples:
            >>> balena.models.application.build_var.get_all_by_application(9020)
            >>> balena.models.application.build_var.get_all_by_application("myorg/myslug")
        """
        return super(BuildEnvVariable, self)._get_all_by_parent(slug_or_uuid_or_id, options)

    def get(self, slug_or_uuid_or_id: Union[str, int], env_var_name: str) -> Optional[str]:
        """
        Get build environment variable.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            env_var_name (str): environment variable name.

        Examples:
            >>> balena.models.application.build_var.get('8deb12','test_env4')
        """
        return super(BuildEnvVariable, self)._get(slug_or_uuid_or_id, env_var_name)

    def set(self, slug_or_uuid_or_id: Union[str, int], env_var_name: str, value: str) -> None:
        """
        Set the value of a specific build environment variable.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            env_var_name (str): environment variable name.
            value (str): environment variable value.

        Examples:
            >>> balena.models.application.build_var.set('8deb12','test_env4', 'testing1')
        """
        super(BuildEnvVariable, self)._set(slug_or_uuid_or_id, env_var_name, value)

    def remove(self, slug_or_uuid_or_id: Union[str, int], key: str) -> None:
        """
        Remove an build environment variable.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)

        Examples:
            >>> balena.models.application.build_var.remove(2184)
        """
        super(BuildEnvVariable, self)._remove(slug_or_uuid_or_id, key)


class ApplicationInvite:
    """
    This class implements application invite model for balena python SDK.

    """

    def __init__(self, pine: PineClient, application: Application, settings: Settings):
        self.__application = application
        self.__pine = pine
        self.__settings = settings
        self.RESOURCE = "invitee__is_invited_to__application"

    def get_all(self, options: AnyObject = {}) -> List[ApplicationInviteType]:
        """
        Get all invites.

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[ApplicationInviteType]: list contains info of invites.

        Examples:
            >>> balena.models.application.invite.get_all()
        """
        return self.__pine.get({"resource": self.RESOURCE, "options": options})

    def get_all_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[ApplicationInviteType]:
        """
        Get all invites by application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            options (AnyObject): extra pine options to use

        Returns:
            List[ApplicationInviteType]: list contains info of invites.

        Examples:
            >>> balena.models.application.invite.get_all_by_application(1681618)
        """
        app = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})
        return self.get_all(
            merge(
                {"$filter": {"is_invited_to__application": app["id"]}},
                options,
            )
        )

    def create(
        self,
        slug_or_uuid_or_id: Union[str, int],
        options: ApplicationInviteOptions,
    ) -> ApplicationInviteType:
        """
        Creates a new invite for an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            options (ApplicationInviteOptions): Application Invite options dict to use.
                invitee (str): the email/balena_username of the invitee.
                role_name (Optional[str]): the role name to be granted to the invitee.
                One of "observer", "developer", "operator". Defaults to "developer"
                message (Optional[str]): the message to send along with the invite.

        Returns:
            dict: application invite.

        Examples:
            >>> balena.models.application.invite.create(1681618, 'invitee@example.org', 'developer', 'Test invite')
        """
        invitee = options.get("invitee")
        if invitee is None:
            raise exceptions.InvalidParameter("options.invitee", None)

        role_name = options.get("roleName", "developer")

        # TODO: paralelize me
        app = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})
        roles = self.__pine.get(
            {
                "resource": "application_membership_role",
                "options": {
                    "$top": 1,
                    "$select": "id",
                    "$filter": {"name": role_name},
                },
            }
        )

        body = {
            "invitee": invitee,
            "is_invited_to__application": app["id"],
            "message": options.get("message"),
        }

        if roles is not None:
            role_id = (roles[0] if len(roles) > 0 else {}).get("id", None)
            if role_id is None:
                raise exceptions.BalenaApplicationMembershipRoleNotFound(role_name)
            body["application_membership_role"] = role_id

        return self.__pine.post({"resource": self.RESOURCE, "body": body})

    def revoke(self, invite_id: int) -> None:
        """
        Revoke an invite.

        Args:
            invite_id (int): application invite id.

        Examples:
            >>> balena.models.application.invite.revoke(5860)
        """
        self.__pine.delete({"resource": self.RESOURCE, "id": invite_id})

    def accept(self, invite_token: str) -> None:
        """
        Accepts an invite.

        Args:
            invite_token (str): invitationToken - invite token.

        Examples:
            >>> balena.models.application.invite.accept("qwerty-invitation-token")
        """
        try:
            request(method="POST", settings=self.__settings, path=f"/user/v1/invitation/{invite_token}")
        except exceptions.RequestError as e:
            if e.status_code == 401:
                raise exceptions.NotLoggedIn()
            raise e


class ApplicationMembership:
    """
    This class implements application membership model for balena python SDK.
    """

    def __init__(self, pine: PineClient, application: Application):
        self.__application = application
        self.__pine = pine
        self.RESOURCE = "user__is_member_of__application"

    def __get_role_id(self, role_name: str) -> Optional[int]:
        role = self.__pine.get(
            {
                "resource": "application_membership_role",
                "id": {"name": role_name},
                "options": {"$select": "id"},
            }
        )

        if role is None:
            raise exceptions.BalenaApplicationMembershipRoleNotFound(role_name)

        return role["id"]

    def get_all(self, options: AnyObject = {}) -> List[ApplicationMembershipType]:
        """
        Get all application memberships.

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[ApplicationMembershipType]: list contains info of application memberships.

        Examples:
            >>> balena.models.application.membership.get_all()
        """

        return self.__pine.get({"resource": self.RESOURCE, "options": options})

    def get(self, membership_id: ResourceKey, options: AnyObject = {}) -> ApplicationMembershipType:
        """
        Get a single application membership.

        Args:
            membership_id (ResourceKey): the id or an object with the unique `user` & `is_member_of__application`
            numeric pair of the membership
            options (AnyObject): extra pine options to use

        Returns:
            ApplicationMembershipType: application membership.

        Examples:
            >>> balena.models.application.membership.get(55074)
            >>> balena.models.application.membership.get({"user": 123, "is_member_of__application": 125})
        """

        if not isinstance(membership_id, int) and not isinstance(membership_id, dict):
            raise exceptions.InvalidParameter("membershipId", membership_id)

        result = self.__pine.get(
            {
                "resource": self.RESOURCE,
                "id": membership_id,  # type: ignore
                "options": options,
            }
        )

        if result is None:
            raise exceptions.ApplicationMembershipNotFound(str(membership_id))

        return result

    def get_all_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[ApplicationMembershipType]:
        """
        Get all memberships by application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            options (AnyObject): extra pine options to use

        Returns:
            list: list contains info of application memberships.

        Examples:
            >>> balena.models.application.membership.get_all_by_application(1681618)
        """
        app = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})
        return self.get_all(
            merge(
                {"$filter": {"is_member_of__application": app["id"]}},
                options,
            )
        )

    def create(
        self,
        slug_or_uuid_or_id: Union[str, int],
        username: str,
        role_name: ApplicationMembershipRoles = "developer",
    ) -> ApplicationMembershipType:
        """
        Creates a new membership for an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            username (str): the username of the balena user that will become a member.
            role_name (Optional[str]): the role name to be granted to the membership.

        Returns:
            ApplicationMembershipType: application membership.

        Examples:
            >>> balena.models.application.membership.create(1681618, 'testuser')
        """

        app = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})
        role_id = self.__get_role_id(role_name)
        body = {
            "username": username,
            "is_member_of__application": app["id"],
            "application_membership_role": role_id,
        }

        return self.__pine.post({"resource": self.RESOURCE, "body": body})

    def change_role(self, membership_id: ResourceKey, role_name: str) -> None:
        """
        Changes the role of an application member.

        Args:
            membership_id (ResourceKey): the id or an object with the unique `user` & `is_member_of__application`
            numeric pair of the membership
            role_name (str): the role name to be granted to the membership.

        Examples:
            >>> balena.models.application.membership.change_role(55074, 'observer')
        """

        role_id = self.__get_role_id(role_name)
        self.__pine.patch(
            {
                "resource": self.RESOURCE,
                "id": membership_id,  # type: ignore
                "body": {"application_membership_role": role_id},
            }
        )

    def remove(self, membership_id: ResourceKey) -> None:
        """
        Remove a membership.

        Args:
            membership_id (ResourceKey): the id or an object with the unique `user` & `is_member_of__application`
        """
        self.__pine.delete({"resource": self.RESOURCE, "id": membership_id})  # type: ignore


from .release import Release  # noqa: E402



from typing import Any, List, Optional, TypedDict, Union
from semver.version import Version

from .. import exceptions
from ..builder import build_from_url
from ..dependent_resource import DependentResource
from ..pine import PineClient
from ..types import AnyObject
from ..types.models import BaseTagType, ReleaseType, ReleaseWithImageDetailsType
from ..utils import is_id, merge
from ..settings import Settings


class ReleaseRawVersionApplicationPair(TypedDict, total=True):
    application: Union[str, int]
    raw_version: str


class Release:
    """
    This class implements release model for balena python SDK.

    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__pine = pine
        self.__application = Application(pine, settings, False)
        self.__settings = settings
        self.tags = ReleaseTag(pine, self, settings)

    def __set(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        body: AnyObject,
    ) -> None:
        release_id = self.get(commit_or_id_or_raw_version, {"$select": "id"})["id"]
        self.__pine.patch({"resource": "release", "id": release_id, "body": body})

    def get(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        options: AnyObject = {},
    ) -> ReleaseType:
        """
        Get a specific release.

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string)
            or id (number) or an object with the unique `application` (number or string) & `rawVersion` (string)
            pair of the release options
            options(AnyObject): extra pine options to use

        Returns:
            ReleaseType: release info.
        """

        if commit_or_id_or_raw_version is None:
            raise exceptions.ReleaseNotFound(commit_or_id_or_raw_version)

        if is_id(commit_or_id_or_raw_version):
            release = self.__pine.get(
                {"resource": "release", "id": commit_or_id_or_raw_version, "options": options}  # type: ignore
            )

            if release is None:
                raise exceptions.ReleaseNotFound(commit_or_id_or_raw_version)
            return release
        else:
            if isinstance(commit_or_id_or_raw_version, dict):
                raw_version = commit_or_id_or_raw_version["raw_version"]
                app = commit_or_id_or_raw_version["application"]
                app_id = self.__application.get(app, {"$select": "id"})["id"]
                dollar_filter = {
                    "raw_version": raw_version,
                    "belongs_to__application": app_id,
                }
            else:
                dollar_filter = {"commit": {"$startswith": commit_or_id_or_raw_version}}
            releases = self.__pine.get(
                {
                    "resource": "release",
                    "options": merge({"$filter": dollar_filter}, options),
                }
            )

            if len(releases) == 0:
                raise exceptions.ReleaseNotFound(str(commit_or_id_or_raw_version))
            if len(releases) > 1:
                raise exceptions.AmbiguousRelease(str(commit_or_id_or_raw_version))
            return releases[0]

    def get_with_image_details(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        image_options: AnyObject = {},
        release_options: AnyObject = {},
    ) -> ReleaseWithImageDetailsType:
        """
        Get a specific release with the details of the images built.

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string)
            image_options (AnyObject): extra pine options to use on image expand
            release_options (AnyObject): extra pine options to use on release expand

        Returns:
            dict: release info.

        Raises:
            ReleaseNotFound: if release couldn't be found.

        """
        base_image_options = {
            "$select": "id",
            "$expand": {"is_a_build_of__service": {"$select": "service_name"}},
        }

        base_release_options = {
            "$expand": {
                "release_image": {"$expand": {"image": merge(base_image_options, image_options)}},
                "is_created_by__user": {"$select": ["id", "username"]},
            }
        }

        release: Any = self.get(
            commit_or_id_or_raw_version,
            merge(base_release_options, release_options),
        )

        images = [ri["image"][0] for ri in release["release_image"]]

        del release["release_image"]
        release["images"] = sorted(
            [
                {
                    **image_data,
                    "service_name": image_data["is_a_build_of__service"][0]["service_name"],
                }
                for image_data in images
                if "is_a_build_of__service" in image_data
            ],
            key=lambda x: x["service_name"],
        )
        created_by_user = release["is_created_by__user"]
        release["user"] = created_by_user[0] if len(created_by_user) > 0 else None

        return release

    def get_all_by_application(self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[ReleaseType]:
        """
        Get all releases from an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            options (AnyObject): extra pine options to use

        Returns:
            List[ReleaseType]: release info.
        """
        app_id = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})["id"]

        return self.__pine.get(
            {
                "resource": "release",
                "options": merge(
                    {
                        "$filter": {"belongs_to__application": app_id},
                        "$orderby": "created_at desc",
                    },
                    options,
                ),
            }
        )

    def get_latest_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> Optional[ReleaseType]:
        """
        Get the latest successful release for an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            options (AnyObject): extra pine options to use

        Returns:
            Optional[ReleaseType]: release info.

        """
        releases = self.get_all_by_application(
            slug_or_uuid_or_id,
            merge({"$top": 1, "$filter": {"status": "success"}}, options),
        )

        if len(releases) == 0:
            return None
        return releases[0]

    def create_from_url(
        self,
        slug_or_uuid_or_id: Union[str, int],
        url: str,
        flatten_tarball: bool = True,
    ) -> int:
        """
        Create a new release built from the source in the provided url.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).
            url (str): a url with a tarball of the project to build.
            flatten_tarball (bool): Should be true when the tarball includes an extra root folder
            with all the content.

        Returns:
            int: release Id.
        """

        app_options = {
            "$select": "app_name",
            "$expand": {"organization": {"$select": "handle"}},
        }
        app = self.__application.get(slug_or_uuid_or_id, app_options)
        return build_from_url(
            app["organization"][0]["handle"],
            app["app_name"],
            url,
            flatten_tarball,
            self.__settings,
        )

    def finalize(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
    ) -> None:
        """
        Finalizes a draft release.

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string)
        """

        self.__set(commit_or_id_or_raw_version, {"is_final": True})

    def set_is_invalidated(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        is_invalidated: bool,
    ) -> None:
        """
        Set the is_invalidated property of a release to True or False.

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string)
            is_invalidated (bool): True for invalidated, False for validated.
        """
        self.__set(commit_or_id_or_raw_version, {"is_invalidated": is_invalidated})

    def set_note(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        note: Optional[str] = None,
    ) -> None:
        """
        Set a note for a release.

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string)
            note (Optional[str]): the note.
        """
        self.__set(commit_or_id_or_raw_version, {"note": note})

    def set_known_issue_list(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        known_issue_list: Optional[str],
    ) -> None:
        """
        Set a known issue list for a release.

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string)
            known_issue_list (Optional[str]): the known issue list.
        """
        self.__set(commit_or_id_or_raw_version, {"known_issue_list": known_issue_list})

    def set_release_version(
        self,
        commit_or_id: Union[str, int],
        semver: str,
    ) -> None:
        """
        Set a direct semver for a given release.

        Args:
            commit_or_id(Union[str, int]): release commit (string) or id (int)
            semver (str): the version to be released, must be a valid semver
        """
        if not Version.is_valid(semver):
            raise exceptions.InvalidParameter("semver", semver)
        self.__set(commit_or_id, {"semver": semver})


class ReleaseTag(DependentResource[BaseTagType]):
    """
    This class implements release tag model for balena python SDK.

    """

    def __init__(self, pine: PineClient, release: Release, settings: Settings):
        self.__release = release
        self.__application = Application(pine, settings, False)
        super(ReleaseTag, self).__init__(
            "release_tag", "tag_key", "release", lambda id: self.__release.get(id, {"$select": "id"})["id"], pine
        )

    def get_all_by_application(self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[BaseTagType]:
        """
        Get all device tags for an application.

        Args:
            slug_or_uuid_or_id (int): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[BaseTagType]: tags list.

        Examples:
            >>> balena.models.release.tag.get_all_by_application(1005160)
        """

        app_id = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})["id"]
        return super(ReleaseTag, self)._get_all(
            merge(
                {
                    "$filter": {
                        "release": {
                            "$any": {
                                "$alias": "r",
                                "$expr": {"r": {"belongs_to__application": app_id}},
                            }
                        }
                    }
                },
                options,
            )
        )

    def get_all_by_release(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        options: AnyObject = {},
    ) -> List[BaseTagType]:
        """
        Get all release tags for a release.

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string) or
            options (AnyObject): extra pine options to use

        Returns:
            List[BaseTagType]: tags list.

        Examples:
            >>> balena.models.release.tags.get_all_by_release(135)
        """

        release_opts = {
            "$select": "id",
            "$expand": {"release_tag": merge({"$orderby": "tag_key asc"}, options)},
        }
        return self.__release.get(commit_or_id_or_raw_version, release_opts)["release_tag"]

    def get_all(self, options: AnyObject = {}) -> List[BaseTagType]:
        """
        Get all release tags.

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[BaseTagType]: tags list.

        Examples:
            >>> balena.models.release.tags.get_all()
        """
        return super(ReleaseTag, self)._get_all(options)

    def set(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        tag_key: str,
        value: str,
    ) -> None:
        """
        Set a release tag (update tag value if it exists).

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string) or
            tag_key (str): tag key.
            value (str): tag value.

        Examples:
            >>> balena.models.release.tags.set(465307, 'releaseTag1', 'Python SDK')
        """
        super(ReleaseTag, self)._set(commit_or_id_or_raw_version, tag_key, value)

    def get(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        tag_key: str,
    ) -> Optional[str]:
        """
        Get a single release tag.

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string) or
            tag_key (str): tag key.

        Examples:
            >>> balena.models.release.tags.get(465307, 'releaseTag1')
        """
        return super(ReleaseTag, self)._get(commit_or_id_or_raw_version, tag_key)

    def remove(
        self,
        commit_or_id_or_raw_version: Union[str, int, ReleaseRawVersionApplicationPair],
        tag_key: str,
    ) -> None:
        """
        Remove a release tag.

        Args:
            commit_or_id_or_raw_version(Union[str, int, ReleaseRawVersionApplicationPair]): release commit (string) or
            tag_key (str): tag key.

        Examples:
            >>> balena.models.release.tags.remove(135, 'releaseTag1')
        """

        super(ReleaseTag, self)._remove(commit_or_id_or_raw_version, tag_key)


from .application import Application  # noqa: E402
